# -*- coding: utf-8 -*-
import numpy as np
from scipy.stats import multivariate_normal

from pgmpy.factors.distributions import BaseDistribution


class GaussianDistribution(BaseDistribution):
    """
    In its most common representation, a multivariate Gaussian distribution
    over X1, X2, ..., Xn is characterized by an n-dimensional mean vector μ,
    and a symmetric n x n covariance matrix Σ.

    This is the base class for its representation.
    """

    def __init__(self, variables, mean, cov):
        """
        Parameters
        ----------
        variables: iterable of any hashable python object
            The variables for which the distribution is defined.

        mean: list, array-like
            1-D array of size n where n is the number of variables.

        cov: n x n, 2-D array like
            n x n dimensional matrix where n is the number of variables.

        Examples
        --------
        >>> import numpy as np
        >>> from pgmpy.factors.distributions import GaussianDistribution as GD
        >>> dis = GD(variables=['x1', 'x2', 'x3'],
        ...          mean=np.array([1, -3, 4]),
        ...          cov=np.array([[4, 2, -2],
        ...                        [2, 5, -5],
        ...                        [-2, -5, 8]]))
        >>> dis.variables
        ['x1', 'x2', 'x3']
        >>> dis.mean
        array([[ 1],
               [-3],
               [4]]))
        >>> dis.cov
        array([[4, 2, -2],
               [2, 5, -5],
               [-2, -5, 8]])
        >>> dis.assignment([0, 0, 0])
        0.0014805631279234139
        """
        no_of_var = len(variables)

        self.variables = variables
        self.mean = np.asarray(np.reshape(mean, (no_of_var, 1)), dtype=float)
        self.covariance = np.asarray(cov, dtype=float)
        self._precision_matrix = None

        if len(mean) != no_of_var:
            raise ValueError(
                "Length of mean_vector must be equal to the", "number of variables."
            )

        if self.covariance.shape != (no_of_var, no_of_var):
            raise ValueError(
                f"The Covariance matrix should be a square matrix  with order equal to the number of variables. Got: {self.covariance.shape}, Expected: {(no_of_var, no_of_var)}"
            )

    @property
    def pdf(self):
        """
        Returns the probability density function(pdf).

        Returns
        -------
        function: The probability density function of the distribution.

        Examples
        --------
        >>> from pgmpy.factors.distributions import GaussianDistribution
        >>> dist = GD(variables=['x1', 'x2', 'x3'],
        ...           mean=[1, -3, 4],
        ...           cov=[[4, 2, -2],
        ...                [2, 5, -5],
        ...                [-2, -5, 8]])
        >>> dist.pdf
        <function pgmpy.factors.distributions.GaussianDistribution.GaussianDistribution.pdf.<locals>.<lambda>>
        >>> dist.pdf([0, 0, 0])
        0.0014805631279234139
        """
        return lambda *args: multivariate_normal.pdf(
            args, self.mean.reshape(1, len(self.variables))[0], self.covariance
        )

    def assignment(self, *x):
        """
        Returns the probability value of the PDF at the given parameter values.

        Parameters
        ----------
        *x: int, float
            The point at which the value of the pdf needs to be computed. The
            number of values passed should be equal to the number of variables
            in the distribution.

        Returns
        -------
        float: float
            The probability value at the point.

        Examples
        --------
        >>> from pgmpy.factors.distributions import GaussianDistribution
        >>> dist = GaussianDistribution(variables=['x1', 'x2'],
        ...                             mean=[0, 0],
        ...                             cov=[[1, 0],
                                             [0, 1]])
        >>> dist.assignment(0, 0)
        0.15915494309189535
        """
        return self.pdf(*x)

    @property
    def precision_matrix(self):
        """
        Returns the precision matrix of the distribution.

        Precision is defined as the inverse of the variance. This method returns
        the inverse matrix of the covariance.

        Examples
        --------
        >>> import numpy as np
        >>> from pgmpy.factors.distributions import GaussianDistribution as GD
        >>> dis = GD(variables=['x1', 'x2', 'x3'],
        ...          mean=[1, -3, 4],
        ...          cov=[[4, 2, -2],
        ...               [2, 5, -5],
        ...               [-2, -5, 8]])
        >>> dis.precision_matrix
        array([[ 0.3125    , -0.125     ,  0.        ],
               [-0.125     ,  0.58333333,  0.33333333],
               [ 0.        ,  0.33333333,  0.33333333]])
        """
        if self._precision_matrix is None:
            self._precision_matrix = np.linalg.inv(self.covariance)
        return self._precision_matrix

    def marginalize(self, variables, inplace=True):
        """
        Modifies the distribution with marginalized values.

        Parameters
        ----------
        variables: iterator over any hashable object.
                List of variables over which marginalization is to be done.

        inplace: boolean
                If inplace=True it will modify the distribution itself,
                else would return a new distribution.

        Returns
        -------
        GaussianDistribution or None :
                if inplace=True (default) returns None
                if inplace=False return a new GaussianDistribution instance

        Examples
        --------
        >>> import numpy as np
        >>> from pgmpy.factors.distributions import GaussianDistribution as GD
        >>> dis = GD(variables=['x1', 'x2', 'x3'],
        ...          mean=[1, -3, 4],
        ...          cov=[[4, 2, -2],
        ...               [2, 5, -5],
        ...               [-2, -5, 8]])
        >>> dis.variables
        ['x1', 'x2', 'x3']
        >>> dis.mean
        array([[ 1],
               [-3],
               [ 4]])
        >>> dis.covariance
        array([[ 4,  2, -2],
               [ 2,  5, -5],
               [-2, -5,  8]])

        >>> dis.marginalize(['x3'])
        dis.variables
        ['x1', 'x2']
        >>> dis.mean
        array([[ 1.],
               [-3.]]))
        >>> dis.covariance
        array([[4., 2.],
               [2., 5.]])
        """
        if not isinstance(variables, list):
            raise TypeError(
                f"variables: Expected type list or array-like, got type {type(variables)}"
            )

        phi = self if inplace else self.copy()

        index_to_keep = [
            self.variables.index(var) for var in self.variables if var not in variables
        ]

        phi.variables = [phi.variables[index] for index in index_to_keep]
        phi.mean = phi.mean[index_to_keep]
        phi.covariance = phi.covariance[np.ix_(index_to_keep, index_to_keep)]
        phi._precision_matrix = None

        if not inplace:
            return phi

    def reduce(self, values, inplace=True):
        """
        Reduces the distribution to the context of the given variable values.

        The formula for the obtained conditional distribution is given by -

        For,
        .. math:: N(X_j | X_i = x_i) ~ N(mu_{j.i} ; sig_{j.i})

        where,
        .. math:: mu_{j.i} = mu_j + sig_{j, i} * {sig_{i, i}^{-1}} * (x_i - mu_i)
        .. math:: sig_{j.i} = sig_{j, j} - sig_{j, i} * {sig_{i, i}^{-1}} * sig_{i, j}

        Parameters
        ----------
        values: list, array-like
            A list of tuples of the form (variable_name, variable_value).

        inplace: boolean
            If inplace=True it will modify the factor itself, else would return
            a new ContinuousFactor object.

        Returns
        -------
        GaussianDistribution or None:
                if inplace=True (default) returns None
                if inplace=False returns a new GaussianDistribution instance.

        Examples
        --------
        >>> import numpy as np
        >>> from pgmpy.factors.distributions import GaussianDistribution as GD
        >>> dis = GD(variables=['x1', 'x2', 'x3'],
        ...             mean=[1, -3, 4],
        ...             cov=[[4, 2, -2],
        ...                  [2, 5, -5],
        ...                  [-2, -5, 8]])
        >>> dis.variables
        ['x1', 'x2', 'x3']
        >>> dis.mean
        array([[ 1.],
               [-3.],
               [ 4.]])
        >>> dis.covariance
        array([[ 4.,  2., -2.],
               [ 2.,  5., -5.],
               [-2., -5.,  8.]])

        >>> dis.reduce([('x1', 7)])
        >>> dis.variables
        ['x2', 'x3']
        >>> dis.mean
        array([[ 0.],
               [ 1.]])
        >>> dis.covariance
        array([[ 4., -4.],
               [-4.,  7.]])

        """
        if not isinstance(values, list):
            raise TypeError(
                f"values: Expected type list or array-like, got type {type(values)}"
            )

        phi = self if inplace else self.copy()

        var_to_reduce = [var for var, value in values]

        # index_to_keep -> j vector
        index_to_keep = [
            self.variables.index(var)
            for var in self.variables
            if var not in var_to_reduce
        ]
        # index_to_reduce -> i vector
        index_to_reduce = [self.variables.index(var) for var in var_to_reduce]

        mu_j = self.mean[index_to_keep]
        mu_i = self.mean[index_to_reduce]
        x_i = np.array([value for var, value in values]).reshape(
            len(index_to_reduce), 1
        )

        sig_i_j = self.covariance[np.ix_(index_to_reduce, index_to_keep)]
        sig_j_i = self.covariance[np.ix_(index_to_keep, index_to_reduce)]
        sig_i_i_inv = np.linalg.inv(
            self.covariance[np.ix_(index_to_reduce, index_to_reduce)]
        )
        sig_j_j = self.covariance[np.ix_(index_to_keep, index_to_keep)]

        phi.variables = [self.variables[index] for index in index_to_keep]
        phi.mean = mu_j + np.linalg.multi_dot([sig_j_i, sig_i_i_inv, x_i - mu_i])
        phi.covariance = sig_j_j - np.linalg.multi_dot([sig_j_i, sig_i_i_inv, sig_i_j])
        phi._precision_matrix = None

        if not inplace:
            return phi

    def normalize(self, inplace=True):
        """
        Normalizes the distribution. In case of a Gaussian Distribution the
        distribution is always normalized, therefore this method doesn't do
        anything and has been implemented only for a consistent API across
        distributions.
        """
        phi = self if inplace else self.copy()

        # The pdf of a Joint Gaussian distribution is always
        # normalized. Hence, no changes.
        if not inplace:
            return phi

    def copy(self):
        """
        Return a copy of the distribution.

        Returns
        -------
        GaussianDistribution: copy of the distribution

        Examples
        --------
        >>> import numpy as np
        >>> from pgmpy.factors.distributions import GaussianDistribution as GD
        >>> gauss_dis = GD(variables=['x1', 'x2', 'x3'],
        ...                mean=[1, -3, 4],
        ...                cov=[[4, 2, -2],
        ...                     [2, 5, -5],
        ...                     [-2, -5, 8]])
        >>> copy_dis = gauss_dis.copy()
        >>> copy_dis.variables
        ['x1', 'x2', 'x3']
        >>> copy_dis.mean
        array([[ 1],
                [-3],
                [ 4]])
        >>> copy_dis.covariance
        array([[ 4,  2, -2],
                [ 2,  5, -5],
                [-2, -5,  8]])
        >>> copy_dis.precision_matrix
        array([[ 0.3125    , -0.125     ,  0.        ],
                [-0.125     ,  0.58333333,  0.33333333],
                [ 0.        ,  0.33333333,  0.33333333]])
        """
        copy_distribution = GaussianDistribution(
            variables=self.variables, mean=self.mean.copy(), cov=self.covariance.copy()
        )
        if self._precision_matrix is not None:
            copy_distribution._precision_matrix = self._precision_matrix.copy()

        return copy_distribution

    def to_canonical_factor(self):
        """
        Returns an equivalent CanonicalDistribution object.

        The formulas for calculating the canonical factor parameters
        for N(μ; Σ) = C(K; h; g) are as follows -

        K = sigma^(-1)
        h = sigma^(-1) * mu
        g = -(0.5) * mu.T * sigma^(-1) * mu -
            log((2*pi)^(n/2) * det(sigma)^(0.5))

        where,
        K,h,g are the canonical factor parameters
        sigma is the covariance_matrix of the distribution,
        mu is the mean_vector of the distribution,
        mu.T is the transpose of the matrix mu,
        and det(sigma) is the determinant of the matrix sigma.

        Example
        -------
        >>> import numpy as np
        >>> from pgmpy.factors.distributions import GaussianDistribution as GD
        >>> dis = GD(variables=['x1', 'x2', 'x3'],
        ...          mean=[1, -3, 4],
        ...          cov=[[4, 2, -2],
        ...               [2, 5, -5],
        ...               [-2, -5, 8]])
        >>> phi = dis.to_canonical_factor()
        >>> phi.variables
        ['x1', 'x2', 'x3']
        >>> phi.K
        array([[0.3125, -0.125, 0.],
               [-0.125, 0.5833, 0.333],
               [     0., 0.333, 0.333]])
        >>> phi.h
        array([[  0.6875],
               [-0.54166],
               [ 0.33333]]))
        >>> phi.g
        -6.51533
        """
        from pgmpy.factors.continuous import CanonicalDistribution

        mu = self.mean
        sigma = self.covariance

        K = self.precision_matrix

        h = np.dot(K, mu)

        g = -(0.5) * np.dot(mu.T, h)[0, 0] - np.log(
            np.power(2 * np.pi, len(self.variables) / 2)
            * np.power(abs(np.linalg.det(sigma)), 0.5)
        )

        return CanonicalDistribution(self.variables, K, h, g)

    def _operate(self, other, operation, inplace=True):
        """
        Gives the CanonicalDistribution operation (product or divide) with
        the other factor.

        Parameters
        ----------
        other: CanonicalDistribution
            The CanonicalDistribution to be multiplied.

        operation: String
            'product' for multiplication operation and
            'divide' for division operation.

        Returns
        -------
        CanonicalDistribution or None:
                        if inplace=True (default) returns None
                        if inplace=False returns a new CanonicalDistribution instance.

        Examples
        --------
        >>> import numpy as np
        >>> from pgmpy.factors.distributions import GaussianDistribution as GD
        >>> dis1 = GD(['x1', 'x2', 'x3'], np.array([[1], [-3], [4]]),
        ...             np.array([[4, 2, -2], [2, 5, -5], [-2, -5, 8]]))
        >>> dis2 = GD(['x3', 'x4'], [1, 2], [[2, 3], [5, 6]])
        >>> dis3 = dis1 * dis2
        >>> dis3.covariance
        array([[ 3.6,  1. , -0.4, -0.6],
               [ 1. ,  2.5, -1. , -1.5],
               [-0.4, -1. ,  1.6,  2.4],
               [-1. , -2.5,  4. ,  4.5]])
        >>> dis3.mean
        array([[ 1.6],
               [-1.5],
               [ 1.6],
               [ 3.5]])
        """
        phi = (
            self.to_canonical_factor()
            ._operate(other.to_canonical_factor(), operation, inplace=False)
            .to_joint_gaussian()
        )

        if not inplace:
            return phi

    def product(self, other, inplace=True):
        """
        TODO: Make it work when using `*` instead of product.

        Returns the product of two gaussian distributions.

        Parameters
        ----------
        other: GaussianDistribution
            The GaussianDistribution to be multiplied.

        inplace: boolean
            If True, modifies the distribution itself, otherwise returns a new
            GaussianDistribution object.

        Returns
        -------
        CanonicalDistribution or None:
                    if inplace=True (default) returns None.
                    if inplace=False returns a new CanonicalDistribution instance.

        Examples
        --------
        >>> import numpy as np
        >>> from pgmpy.factors.distributions import GaussianDistribution as GD
        >>> dis1 = GD(['x1', 'x2', 'x3'], np.array([[1], [-3], [4]]),
        ...            np.array([[4, 2, -2], [2, 5, -5], [-2, -5, 8]]))
        >>> dis2 = GD(['x3', 'x4'], [1, 2], [[2, 3], [5, 6]])
        >>> dis3 = dis1.product(dis2, inplace=False)
        >>> dis3.covariance
        array([[ 3.6,  1. , -0.4, -0.6],
               [ 1. ,  2.5, -1. , -1.5],
               [-0.4, -1. ,  1.6,  2.4],
               [-1. , -2.5,  4. ,  4.5]])
        >>> dis3.mean
        array([[ 1.6],
               [-1.5],
               [ 1.6],
               [ 3.5]])
        """
        return self._operate(other, operation="product", inplace=inplace)

    def divide(self, other, inplace=True):
        """
        Returns the division of two gaussian distributions.

        Parameters
        ----------
        other: GaussianDistribution
            The GaussianDistribution to be divided.

        inplace: boolean
            If True, modifies the distribution itself, otherwise returns a new
            GaussianDistribution object.

        Returns
        -------
        CanonicalDistribution or None:
                    if inplace=True (default) returns None.
                    if inplace=False returns a new CanonicalDistribution instance.

        Examples
        --------
        >>> import numpy as np
        >>> from pgmpy.factors.distributions import GaussianDistribution as GD
        >>> dis1 = GD(['x1', 'x2', 'x3'], np.array([[1], [-3], [4]]),
        ...            np.array([[4, 2, -2], [2, 5, -5], [-2, -5, 8]]))
        >>> dis2 = GD(['x3', 'x4'], [1, 2], [[2, 3], [5, 6]])
        >>> dis3 = dis1.divide(dis2, inplace=False)
        >>> dis3.covariance
        array([[ 3.6,  1. , -0.4, -0.6],
               [ 1. ,  2.5, -1. , -1.5],
               [-0.4, -1. ,  1.6,  2.4],
               [-1. , -2.5,  4. ,  4.5]])
        >>> dis3.mean
        array([[ 1.6],
               [-1.5],
               [ 1.6],
               [ 3.5]])
        """
        return self._operate(other, operation="divide", inplace=inplace)

    def __repr__(self):
        return (
            f"GaussianDistribution representing N({self.variables}) at {hex(id(self))}"
        )

    def __mul__(self, other):
        return self.product(other, inplace=False)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        return self.divide(other, inplace=False)

    __div__ = __truediv__

    def __eq__(self, other):
        if not (
            isinstance(self, GaussianDistribution)
            and isinstance(self, GaussianDistribution)
        ):
            return False

        elif set(self.scope()) != set(other.scope()):
            return False

        else:
            # Computing transform_index to be able to easily have variables in same order.
            transform_index = [other.index(var) for var in self.variables]

            if not np.allclose(self.mean, other.mean[transform_index]):
                return False
            else:
                mid_cov = other.covariance[transform_index, :]
                transform_cov = mid_cov[:, transform_index]
                if not np.allclose(self.covariance, transform_cov):
                    return False
        return True



# -*- coding: utf-8 -*-

import itertools


class Independencies(object):
    """
    Base class for independencies.
    independencies class represents a set of Conditional Independence
    assertions (eg: "X is independent of Y given Z" where X, Y and Z
    are random variables) or Independence assertions (eg: "X is
    independent of Y" where X and Y are random variables).
    Initialize the independencies Class with Conditional Independence
    assertions or Independence assertions.

    Parameters
    ----------
    assertions: Lists or Tuples
            Each assertion is a list or tuple of the form: [event1,
            event2 and event3]
            eg: assertion ['X', 'Y', 'Z'] would be X is independent
            of Y given Z.

    Examples
    --------
    Creating an independencies object with one independence assertion:
    Random Variable X is independent of Y

    >>> independencies = independencies(['X', 'Y'])

    Creating an independencies object with three conditional
    independence assertions:
    First assertion is Random Variable X is independent of Y given Z.

    >>> independencies = independencies(['X', 'Y', 'Z'],
    ...             ['a', ['b', 'c'], 'd'],
    ...             ['l', ['m', 'n'], 'o'])

    Public Methods
    --------------
    add_assertions
    get_assertions
    get_factorized_product
    closure
    entails
    is_equivalent
    """

    def __init__(self, *assertions):
        self.independencies = []
        self.add_assertions(*assertions)

    def __str__(self):
        string = "\n".join([str(assertion) for assertion in self.independencies])
        return string

    __repr__ = __str__

    def __eq__(self, other):
        if not isinstance(other, Independencies):
            return False
        return all(
            independency in other.get_assertions()
            for independency in self.get_assertions()
        ) and all(
            independency in self.get_assertions()
            for independency in other.get_assertions()
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def contains(self, assertion):
        """
        Returns `True` if `assertion` is contained in this `Independencies`-object,
        otherwise `False`.

        Parameters
        ----------
        assertion: IndependenceAssertion()-object

        Examples
        --------
        >>> from pgmpy.independencies import Independencies, IndependenceAssertion
        >>> ind = Independencies(['A', 'B', ['C', 'D']])
        >>> IndependenceAssertion('A', 'B', ['C', 'D']) in ind
        True
        >>> # does not depend on variable order:
        >>> IndependenceAssertion('B', 'A', ['D', 'C']) in ind
        True
        >>> # but does not check entailment:
        >>> IndependenceAssertion('X', 'Y', 'Z') in Independencies(['X', 'Y'])
        False
        """
        if not isinstance(assertion, IndependenceAssertion):
            raise TypeError(
                f"' in <Independencies()>' requires IndependenceAssertion as left operand, not {type(assertion)}"
            )

        return assertion in self.get_assertions()

    __contains__ = contains

    def get_all_variables(self):
        """
        Returns a set of all the variables in all the independence assertions.
        """
        return frozenset().union(*[ind.all_vars for ind in self.independencies])

    def get_assertions(self):
        """
        Returns the independencies object which is a set of IndependenceAssertion objects.

        Examples
        --------
        >>> from pgmpy.independencies import Independencies
        >>> independencies = Independencies(['X', 'Y', 'Z'])
        >>> independencies.get_assertions()
        """
        return self.independencies

    def add_assertions(self, *assertions):
        """
        Adds assertions to independencies.

        Parameters
        ----------
        assertions: Lists or Tuples
                Each assertion is a list or tuple of variable, independent_of and given.

        Examples
        --------
        >>> from pgmpy.independencies import Independencies
        >>> independencies = Independencies()
        >>> independencies.add_assertions(['X', 'Y', 'Z'])
        >>> independencies.add_assertions(['a', ['b', 'c'], 'd'])
        """
        for assertion in assertions:
            if isinstance(assertion, IndependenceAssertion):
                self.independencies.append(assertion)
            else:
                try:
                    self.independencies.append(
                        IndependenceAssertion(assertion[0], assertion[1], assertion[2])
                    )
                except IndexError:
                    self.independencies.append(
                        IndependenceAssertion(assertion[0], assertion[1])
                    )

    def closure(self):
        """
        Returns a new `Independencies()`-object that additionally contains those `IndependenceAssertions`
        that are implied by the current independencies (using with the `semi-graphoid axioms
        <https://en.wikipedia.org/w/index.php?title=Conditional_independence&oldid=708760689#Rules_of_conditional_independence>`_;
        see (Pearl, 1989, `Conditional Independence and its representations
        <http://www.cs.technion.ac.il/~dang/journal_papers/pearl1989conditional.pdf>`_)).

        Might be very slow if more than six variables are involved.

        Examples
        --------
        >>> from pgmpy.independencies import Independencies
        >>> ind1 = Independencies(('A', ['B', 'C'], 'D'))
        >>> ind1.closure()
        (A \u27C2 B | D, C)
        (A \u27C2 B, C | D)
        (A \u27C2 B | D)
        (A \u27C2 C | D, B)
        (A \u27C2 C | D)

        >>> ind2 = Independencies(('W', ['X', 'Y', 'Z']))
        >>> ind2.closure()
        (W \u27C2 Y)
        (W \u27C2 Y | X)
        (W \u27C2 Z | Y)
        (W \u27C2 Z, X, Y)
        (W \u27C2 Z)
        (W \u27C2 Z, X)
        (W \u27C2 X, Y)
        (W \u27C2 Z | X)
        (W \u27C2 Z, Y | X)
        [..]
        """

        def single_var(var):
            "Checks if var represents a single variable"
            if not hasattr(var, "__iter__"):
                return True
            else:
                return len(var) == 1

        def sg0(ind):
            "Symmetry rule: 'X ⟂ Y | Z' -> 'Y ⟂ X | Z'"
            return IndependenceAssertion(ind.event2, ind.event1, ind.event3)

        # since X⟂Y|Z == Y⟂X|Z in pgmpy, sg0 (symmetry) is not used as an axiom/rule.
        # instead we use a decorator for the other axioms to apply them on both sides
        def apply_left_and_right(func):
            def symmetric_func(*args):
                if len(args) == 1:
                    return func(args[0]) + func(sg0(args[0]))
                if len(args) == 2:
                    return (
                        func(*args)
                        + func(args[0], sg0(args[1]))
                        + func(sg0(args[0]), args[1])
                        + func(sg0(args[0]), sg0(args[1]))
                    )

            return symmetric_func

        @apply_left_and_right
        def sg1(ind):
            "Decomposition rule: 'X ⟂ Y,W | Z' -> 'X ⟂ Y | Z', 'X ⟂ W | Z'"
            if single_var(ind.event2):
                return []
            else:
                return [
                    IndependenceAssertion(ind.event1, ind.event2 - {elem}, ind.event3)
                    for elem in ind.event2
                ]

        @apply_left_and_right
        def sg2(ind):
            "Weak Union rule: 'X ⟂ Y,W | Z' -> 'X ⟂ Y | W,Z', 'X ⟂ W | Y,Z'"
            if single_var(ind.event2):
                return []
            else:
                return [
                    IndependenceAssertion(
                        ind.event1, ind.event2 - {elem}, {elem} | ind.event3
                    )
                    for elem in ind.event2
                ]

        @apply_left_and_right
        def sg3(ind1, ind2):
            "Contraction rule: 'X ⟂ W | Y,Z' & 'X ⟂ Y | Z' -> 'X ⟂ W,Y | Z'"
            if ind1.event1 != ind2.event1:
                return []

            Y = ind2.event2
            Z = ind2.event3
            Y_Z = ind1.event3
            if Y < Y_Z and Z < Y_Z and Y.isdisjoint(Z):
                return [IndependenceAssertion(ind1.event1, ind1.event2 | Y, Z)]
            else:
                return []

        # apply semi-graphoid axioms as long as new independencies are found.
        all_independencies = set()
        new_inds = set(self.independencies)

        while new_inds:
            new_pairs = (
                set(itertools.permutations(new_inds, 2))
                | set(itertools.product(new_inds, all_independencies))
                | set(itertools.product(all_independencies, new_inds))
            )

            all_independencies |= new_inds
            new_inds = set(
                sum(
                    [sg1(ind) for ind in new_inds]
                    + [sg2(ind) for ind in new_inds]
                    + [sg3(*inds) for inds in new_pairs],
                    [],
                )
            )
            new_inds -= all_independencies

        return Independencies(*list(all_independencies))

    def entails(self, entailed_independencies):
        """
        Returns `True` if the `entailed_independencies` are implied by this `Independencies`-object, otherwise `False`.
        Entailment is checked using the semi-graphoid axioms.

        Might be very slow if more than six variables are involved.

        Parameters
        ----------
        entailed_independencies: Independencies()-object

        Examples
        --------
        >>> from pgmpy.independencies import Independencies
        >>> ind1 = Independencies([['A', 'B'], ['C', 'D'], 'E'])
        >>> ind2 = Independencies(['A', 'C', 'E'])
        >>> ind1.entails(ind2)
        True
        >>> ind2.entails(ind1)
        False
        """
        if not isinstance(entailed_independencies, Independencies):
            return False

        implications = self.closure().get_assertions()
        return all(
            ind in implications for ind in entailed_independencies.get_assertions()
        )

    def is_equivalent(self, other):
        """
        Returns True if the two Independencies-objects are equivalent, otherwise False.
        (i.e. any Bayesian Network that satisfies the one set
        of conditional independencies also satisfies the other).

        Might be very slow if more than six variables are involved.

        Parameters
        ----------
        other: Independencies()-object

        Examples
        --------
        >>> from pgmpy.independencies import Independencies
        >>> ind1 = Independencies(['X', ['Y', 'W'], 'Z'])
        >>> ind2 = Independencies(['X', 'Y', 'Z'], ['X', 'W', 'Z'])
        >>> ind3 = Independencies(['X', 'Y', 'Z'], ['X', 'W', 'Z'], ['X', 'Y', ['W','Z']])
        >>> ind1.is_equivalent(ind2)
        False
        >>> ind1.is_equivalent(ind3)
        True
        """
        return self.entails(other) and other.entails(self)

        # TODO: write reduce function.

    def reduce(self):
        """
        Add function to remove duplicate Independence Assertions
        """
        pass

    def latex_string(self):
        """
        Returns a list of string.
        Each string represents the IndependenceAssertion in latex.
        """
        return [assertion.latex_string() for assertion in self.get_assertions()]

    def get_factorized_product(self, random_variables=None, latex=False):
        # TODO: Write this whole function
        #
        # The problem right now is that the factorized product for all
        # P(A, B, C), P(B, A, C) etc should be same but on solving normally
        # we get different results which have to be simplified to a simpler
        # form. How to do that ??? and also how to decide which is the most
        # simplified form???
        #
        pass


class IndependenceAssertion(object):
    r"""
    Represents Conditional Independence or Independence assertion.

    Each assertion has 3 attributes: event1, event2, event3.
    The attributes for

    .. math:: U \perp X, Y | Z

    is read as: Random Variable U is independent of X and Y given Z would be:

    event1 = {U}

    event2 = {X, Y}

    event3 = {Z}

    Parameters
    ----------
    event1: String or List of strings
            Random Variable which is independent.

    event2: String or list of strings.
            Random Variables from which event1 is independent

    event3: String or list of strings.
            Random Variables given which event1 is independent of event2.

    Examples
    --------
    >>> from pgmpy.independencies import IndependenceAssertion
    >>> assertion = IndependenceAssertion('U', 'X')
    >>> assertion = IndependenceAssertion('U', ['X', 'Y'])
    >>> assertion = IndependenceAssertion('U', ['X', 'Y'], 'Z')
    >>> assertion = IndependenceAssertion(['U', 'V'], ['X', 'Y'], ['Z', 'A'])


    Public Methods
    --------------
    get_assertion
    """

    def __init__(self, event1=[], event2=[], event3=[]):
        r"""
        Initialize an IndependenceAssertion object with event1, event2 and event3 attributes.

                    event2
                    ^
        event1     /   event3
           ^      /     ^
           |     /      |
          (U || X, Y | Z) read as Random variable U is independent of X and Y given Z.
            ---
        """
        if event1 and not event2:
            raise ValueError("event2 needs to be specified")
        if any([event2, event3]) and not event1:
            raise ValueError("event1 needs to be specified")
        if event3 and not all([event1, event2]):
            raise ValueError(
                "event1" if not event1 else "event2" + " needs to be specified"
            )

        self.event1 = frozenset(self._return_list_if_not_collection(event1))
        self.event2 = frozenset(self._return_list_if_not_collection(event2))
        self.event3 = frozenset(self._return_list_if_not_collection(event3))
        self.all_vars = frozenset().union(self.event1, self.event2, self.event3)

    def __str__(self):
        if self.event3:
            return "({event1} \u27C2 {event2} | {event3})".format(
                event1=", ".join([str(e) for e in self.event1]),
                event2=", ".join([str(e) for e in self.event2]),
                event3=", ".join([str(e) for e in self.event3]),
            )
        else:
            return "({event1} \u27C2 {event2})".format(
                event1=", ".join([str(e) for e in self.event1]),
                event2=", ".join([str(e) for e in self.event2]),
            )

    __repr__ = __str__

    def __eq__(self, other):
        if not isinstance(other, IndependenceAssertion):
            return False
        return (self.event1, self.event2, self.event3) == other.get_assertion() or (
            self.event2,
            self.event1,
            self.event3,
        ) == other.get_assertion()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash((frozenset((self.event1, self.event2)), self.event3))

    @staticmethod
    def _return_list_if_not_collection(event):
        """
        If variable is a string returns a list containing variable.
        Else returns variable itself.
        """
        if isinstance(event, str):
            return [event]
        else:
            return event

    def get_assertion(self):
        """
        Returns a tuple of the attributes: variable, independent_of, given.

        Examples
        --------
        >>> from pgmpy.independencies import IndependenceAssertion
        >>> asser = IndependenceAssertion('X', 'Y', 'Z')
        >>> asser.get_assertion()
        """
        return self.event1, self.event2, self.event3

    def latex_string(self):
        if len(self.event3) == 0:
            return r"{event1} \perp {event2}".format(
                event1=", ".join([str(e) for e in self.event1]),
                event2=", ".join([str(e) for e in self.event2]),
            )
        else:
            return r"{event1} \perp {event2} \mid {event3}".format(
                event1=", ".join([str(e) for e in self.event1]),
                event2=", ".join([str(e) for e in self.event2]),
                event3=", ".join([str(e) for e in self.event3]),
            )



import os
import unittest

import numpy as np
import networkx as nx
import numpy.testing as npt

from pgmpy.models import SEM, SEMGraph, SEMAlg


class TestSEM(unittest.TestCase):
    def test_from_graph(self):
        self.demo = SEM.from_graph(
            ebunch=[
                ("xi1", "x1"),
                ("xi1", "x2"),
                ("xi1", "x3"),
                ("xi1", "eta1"),
                ("eta1", "y1"),
                ("eta1", "y2"),
                ("eta1", "y3"),
                ("eta1", "y4"),
                ("eta1", "eta2"),
                ("xi1", "eta2"),
                ("eta2", "y5"),
                ("eta2", "y6"),
                ("eta2", "y7"),
                ("eta2", "y8"),
            ],
            latents=["xi1", "eta1", "eta2"],
            err_corr=[
                ("y1", "y5"),
                ("y2", "y6"),
                ("y2", "y4"),
                ("y3", "y7"),
                ("y4", "y8"),
                ("y6", "y8"),
            ],
        )

        self.assertSetEqual(self.demo.latents, {"xi1", "eta1", "eta2"})
        self.assertSetEqual(
            self.demo.observed,
            {"x1", "x2", "x3", "y1", "y2", "y3", "y4", "y5", "y6", "y7", "y8"},
        )
        self.assertListEqual(
            sorted(self.demo.graph.nodes()),
            [
                "eta1",
                "eta2",
                "x1",
                "x2",
                "x3",
                "xi1",
                "y1",
                "y2",
                "y3",
                "y4",
                "y5",
                "y6",
                "y7",
                "y8",
            ],
        )
        self.assertListEqual(
            sorted(self.demo.graph.edges()),
            sorted(
                [
                    ("eta1", "eta2"),
                    ("eta1", "y1"),
                    ("eta1", "y2"),
                    ("eta1", "y3"),
                    ("eta1", "y4"),
                    ("eta2", "y5"),
                    ("eta2", "y6"),
                    ("eta2", "y7"),
                    ("eta2", "y8"),
                    ("xi1", "eta1"),
                    ("xi1", "eta2"),
                    ("xi1", "x1"),
                    ("xi1", "x2"),
                    ("xi1", "x3"),
                ]
            ),
        )

        self.assertDictEqual(self.demo.graph.edges[("xi1", "x1")], {"weight": np.NaN})
        self.assertDictEqual(self.demo.graph.edges[("xi1", "x2")], {"weight": np.NaN})
        self.assertDictEqual(self.demo.graph.edges[("xi1", "x3")], {"weight": np.NaN})
        self.assertDictEqual(self.demo.graph.edges[("xi1", "eta1")], {"weight": np.NaN})
        self.assertDictEqual(self.demo.graph.edges[("eta1", "y1")], {"weight": np.NaN})
        self.assertDictEqual(self.demo.graph.edges[("eta1", "y2")], {"weight": np.NaN})
        self.assertDictEqual(self.demo.graph.edges[("eta1", "y3")], {"weight": np.NaN})
        self.assertDictEqual(self.demo.graph.edges[("eta1", "y4")], {"weight": np.NaN})
        self.assertDictEqual(
            self.demo.graph.edges[("eta1", "eta2")], {"weight": np.NaN}
        )
        self.assertDictEqual(self.demo.graph.edges[("xi1", "eta2")], {"weight": np.NaN})
        self.assertDictEqual(self.demo.graph.edges[("eta2", "y5")], {"weight": np.NaN})
        self.assertDictEqual(self.demo.graph.edges[("eta2", "y6")], {"weight": np.NaN})
        self.assertDictEqual(self.demo.graph.edges[("eta2", "y7")], {"weight": np.NaN})
        self.assertDictEqual(self.demo.graph.edges[("eta2", "y8")], {"weight": np.NaN})

        npt.assert_equal(
            nx.to_numpy_matrix(
                self.demo.err_graph,
                nodelist=sorted(self.demo.err_graph.nodes()),
                weight=None,
            ),
            np.array(
                [
                    [
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                    ],
                    [
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                    ],
                    [
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                    ],
                    [
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                    ],
                    [
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                    ],
                    [
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                    ],
                    [
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        1.0,
                        0.0,
                        0.0,
                        0.0,
                    ],
                    [
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        1.0,
                        0.0,
                        1.0,
                        0.0,
                        0.0,
                    ],
                    [
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        1.0,
                        0.0,
                    ],
                    [
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        1.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        1.0,
                    ],
                    [
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        1.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                    ],
                    [
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        1.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        1.0,
                    ],
                    [
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        1.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                    ],
                    [
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        1.0,
                        0.0,
                        1.0,
                        0.0,
                        0.0,
                    ],
                ]
            ),
        )

        for edge in self.demo.err_graph.edges():
            self.assertDictEqual(self.demo.err_graph.edges[edge], {"weight": np.NaN})
        for node in self.demo.err_graph.nodes():
            self.assertDictEqual(self.demo.err_graph.nodes[node], {"weight": np.NaN})

    def test_from_lavaan(self):
        model_str = """# %load model.lav
                       # measurement model
                         ind60 =~ x1 + x2 + x3
                         dem60 =~ y1 + y2 + y3 + y4
                         dem65 =~ y5 + y6 + y7 + y8
                       # regressions
                         dem60 ~ ind60
                         dem65 ~ ind60 + dem60
                       # residual correlations
                         y1 ~~ y5
                         y2 ~~ y4 + y6
                         y3 ~~ y7
                         y4 ~~ y8
                         y6 ~~ y8
                       """
        model_from_str = SEM.from_lavaan(string=model_str)

        with open("test_model.lav", "w") as f:
            f.write(model_str)
        model_from_file = SEM.from_lavaan(filename="test_model.lav")
        os.remove("test_model.lav")

        expected_edges = set(
            [
                ("ind60", "x1"),
                ("ind60", "x2"),
                ("ind60", "x3"),
                ("ind60", "dem60"),
                ("ind60", "dem65"),
                ("dem60", "dem65"),
                ("dem60", "y1"),
                ("dem60", "y2"),
                ("dem60", "y3"),
                ("dem60", "y4"),
                ("dem65", "y5"),
                ("dem65", "y6"),
                ("dem65", "y7"),
                ("dem65", "y8"),
            ]
        )

        # Undirected Graph, needs to handle when edges returned in reverse.
        expected_err_edges = set(
            [
                ("y1", "y5"),
                ("y5", "y1"),
                ("y2", "y6"),
                ("y6", "y2"),
                ("y2", "y4"),
                ("y4", "y2"),
                ("y3", "y7"),
                ("y7", "y3"),
                ("y4", "y8"),
                ("y8", "y4"),
                ("y6", "y8"),
                ("y8", "y6"),
            ]
        )

        expected_latents = set(["dem60", "dem65", "ind60"])
        self.assertEqual(set(model_from_str.graph.edges()), expected_edges)
        self.assertEqual(set(model_from_file.graph.edges()), expected_edges)
        self.assertFalse(set(model_from_str.err_graph.edges()) - expected_err_edges)
        self.assertFalse(set(model_from_file.err_graph.edges()) - expected_err_edges)
        self.assertEqual(set(model_from_str.latents), expected_latents)
        self.assertEqual(set(model_from_file.latents), expected_latents)

    def test_from_lisrel(self):
        pass  # TODO: Add this test when done writing the tests for SEMAlg

    def test_from_ram(self):
        pass  # TODO: Add this.


class TestSEMGraph(unittest.TestCase):
    def setUp(self):
        self.demo = SEMGraph(
            ebunch=[
                ("xi1", "x1"),
                ("xi1", "x2"),
                ("xi1", "x3"),
                ("xi1", "eta1"),
                ("eta1", "y1"),
                ("eta1", "y2"),
                ("eta1", "y3"),
                ("eta1", "y4"),
                ("eta1", "eta2"),
                ("xi1", "eta2"),
                ("eta2", "y5"),
                ("eta2", "y6"),
                ("eta2", "y7"),
                ("eta2", "y8"),
            ],
            latents=["xi1", "eta1", "eta2"],
            err_corr=[
                ("y1", "y5"),
                ("y2", "y6"),
                ("y2", "y4"),
                ("y3", "y7"),
                ("y4", "y8"),
                ("y6", "y8"),
            ],
        )

        self.union = SEMGraph(
            ebunch=[
                ("yrsmill", "unionsen"),
                ("age", "laboract"),
                ("age", "deferenc"),
                ("deferenc", "laboract"),
                ("deferenc", "unionsen"),
                ("laboract", "unionsen"),
            ],
            latents=[],
            err_corr=[("yrsmill", "age")],
        )

        self.demo_params = SEMGraph(
            ebunch=[
                ("xi1", "x1", 0.4),
                ("xi1", "x2", 0.5),
                ("xi1", "x3", 0.6),
                ("xi1", "eta1", 0.3),
                ("eta1", "y1", 1.1),
                ("eta1", "y2", 1.2),
                ("eta1", "y3", 1.3),
                ("eta1", "y4", 1.4),
                ("eta1", "eta2", 0.1),
                ("xi1", "eta2", 0.2),
                ("eta2", "y5", 0.7),
                ("eta2", "y6", 0.8),
                ("eta2", "y7", 0.9),
                ("eta2", "y8", 1.0),
            ],
            latents=["xi1", "eta1", "eta2"],
            err_corr=[
                ("y1", "y5", 1.5),
                ("y2", "y6", 1.6),
                ("y2", "y4", 1.9),
                ("y3", "y7", 1.7),
                ("y4", "y8", 1.8),
                ("y6", "y8", 2.0),
            ],
            err_var={
                "y1": 2.1,
                "y2": 2.2,
                "y3": 2.3,
                "y4": 2.4,
                "y5": 2.5,
                "y6": 2.6,
                "y7": 2.7,
                "y8": 2.8,
                "x1": 3.1,
                "x2": 3.2,
                "x3": 3.3,
                "eta1": 2.9,
                "eta2": 3.0,
                "xi1": 3.4,
            },
        )

        self.custom = SEMGraph(
            ebunch=[
                ("xi1", "eta1"),
                ("xi1", "y1"),
                ("xi1", "y4"),
                ("xi1", "x1"),
                ("xi1", "x2"),
                ("y4", "y1"),
                ("y1", "eta2"),
                ("eta2", "y5"),
                ("y1", "eta1"),
                ("eta1", "y2"),
                ("eta1", "y3"),
            ],
            latents=["xi1", "eta1", "eta2"],
            err_corr=[("y1", "y2"), ("y2", "y3")],
            err_var={},
        )

    def test_demo_init(self):
        self.assertSetEqual(self.demo.latents, {"xi1", "eta1", "eta2"})
        self.assertSetEqual(
            self.demo.observed,
            {"x1", "x2", "x3", "y1", "y2", "y3", "y4", "y5", "y6", "y7", "y8"},
        )
        self.assertListEqual(
            sorted(self.demo.graph.nodes()),
            [
                "eta1",
                "eta2",
                "x1",
                "x2",
                "x3",
                "xi1",
                "y1",
                "y2",
                "y3",
                "y4",
                "y5",
                "y6",
                "y7",
                "y8",
            ],
        )
        self.assertListEqual(
            sorted(self.demo.graph.edges()),
            sorted(
                [
                    ("eta1", "eta2"),
                    ("eta1", "y1"),
                    ("eta1", "y2"),
                    ("eta1", "y3"),
                    ("eta1", "y4"),
                    ("eta2", "y5"),
                    ("eta2", "y6"),
                    ("eta2", "y7"),
                    ("eta2", "y8"),
                    ("xi1", "eta1"),
                    ("xi1", "eta2"),
                    ("xi1", "x1"),
                    ("xi1", "x2"),
                    ("xi1", "x3"),
                ]
            ),
        )

        self.assertDictEqual(self.demo.graph.edges[("xi1", "x1")], {"weight": np.NaN})
        self.assertDictEqual(self.demo.graph.edges[("xi1", "x2")], {"weight": np.NaN})
        self.assertDictEqual(self.demo.graph.edges[("xi1", "x3")], {"weight": np.NaN})
        self.assertDictEqual(self.demo.graph.edges[("xi1", "eta1")], {"weight": np.NaN})
        self.assertDictEqual(self.demo.graph.edges[("eta1", "y1")], {"weight": np.NaN})
        self.assertDictEqual(self.demo.graph.edges[("eta1", "y2")], {"weight": np.NaN})
        self.assertDictEqual(self.demo.graph.edges[("eta1", "y3")], {"weight": np.NaN})
        self.assertDictEqual(self.demo.graph.edges[("eta1", "y4")], {"weight": np.NaN})
        self.assertDictEqual(
            self.demo.graph.edges[("eta1", "eta2")], {"weight": np.NaN}
        )
        self.assertDictEqual(self.demo.graph.edges[("xi1", "eta2")], {"weight": np.NaN})
        self.assertDictEqual(self.demo.graph.edges[("eta2", "y5")], {"weight": np.NaN})
        self.assertDictEqual(self.demo.graph.edges[("eta2", "y6")], {"weight": np.NaN})
        self.assertDictEqual(self.demo.graph.edges[("eta2", "y7")], {"weight": np.NaN})
        self.assertDictEqual(self.demo.graph.edges[("eta2", "y8")], {"weight": np.NaN})

        npt.assert_equal(
            nx.to_numpy_matrix(
                self.demo.err_graph,
                nodelist=sorted(self.demo.err_graph.nodes()),
                weight=None,
            ),
            np.array(
                [
                    [
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                    ],
                    [
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                    ],
                    [
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                    ],
                    [
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                    ],
                    [
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                    ],
                    [
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                    ],
                    [
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        1.0,
                        0.0,
                        0.0,
                        0.0,
                    ],
                    [
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        1.0,
                        0.0,
                        1.0,
                        0.0,
                        0.0,
                    ],
                    [
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        1.0,
                        0.0,
                    ],
                    [
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        1.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        1.0,
                    ],
                    [
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        1.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                    ],
                    [
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        1.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        1.0,
                    ],
                    [
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        1.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                    ],
                    [
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        0.0,
                        1.0,
                        0.0,
                        1.0,
                        0.0,
                        0.0,
                    ],
                ]
            ),
        )

        for edge in self.demo.err_graph.edges():
            self.assertDictEqual(self.demo.err_graph.edges[edge], {"weight": np.NaN})
        for node in self.demo.err_graph.nodes():
            self.assertDictEqual(self.demo.err_graph.nodes[node], {"weight": np.NaN})

    def test_union_init(self):
        self.assertSetEqual(self.union.latents, set())
        self.assertSetEqual(
            self.union.observed, {"yrsmill", "unionsen", "age", "laboract", "deferenc"}
        )

        self.assertListEqual(
            sorted(self.union.graph.nodes()),
            sorted(["yrsmill", "unionsen", "age", "laboract", "deferenc"]),
        )
        self.assertListEqual(
            sorted(self.union.graph.edges()),
            sorted(
                [
                    ("yrsmill", "unionsen"),
                    ("age", "laboract"),
                    ("age", "deferenc"),
                    ("deferenc", "laboract"),
                    ("deferenc", "unionsen"),
                    ("laboract", "unionsen"),
                ]
            ),
        )

        self.assertDictEqual(
            self.union.graph.edges[("yrsmill", "unionsen")], {"weight": np.NaN}
        )
        self.assertDictEqual(
            self.union.graph.edges[("age", "laboract")], {"weight": np.NaN}
        )
        self.assertDictEqual(
            self.union.graph.edges[("age", "deferenc")], {"weight": np.NaN}
        )
        self.assertDictEqual(
            self.union.graph.edges[("deferenc", "laboract")], {"weight": np.NaN}
        )
        self.assertDictEqual(
            self.union.graph.edges[("deferenc", "unionsen")], {"weight": np.NaN}
        )
        self.assertDictEqual(
            self.union.graph.edges[("laboract", "unionsen")], {"weight": np.NaN}
        )

        npt.assert_equal(
            nx.to_numpy_matrix(
                self.union.err_graph,
                nodelist=sorted(self.union.err_graph.nodes()),
                weight=None,
            ),
            np.array(
                [
                    [0.0, 0.0, 0.0, 0.0, 1.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                    [0.0, 0.0, 0.0, 0.0, 0.0],
                    [1.0, 0.0, 0.0, 0.0, 0.0],
                ]
            ),
        )

        for edge in self.union.err_graph.edges():
            self.assertDictEqual(self.union.err_graph.edges[edge], {"weight": np.NaN})
        for node in self.union.err_graph.nodes():
            self.assertDictEqual(self.union.err_graph.nodes[node], {"weight": np.NaN})

    def test_demo_param_init(self):
        self.assertDictEqual(
            self.demo_params.graph.edges[("xi1", "x1")], {"weight": 0.4}
        )
        self.assertDictEqual(
            self.demo_params.graph.edges[("xi1", "x2")], {"weight": 0.5}
        )
        self.assertDictEqual(
            self.demo_params.graph.edges[("xi1", "x3")], {"weight": 0.6}
        )
        self.assertDictEqual(
            self.demo_params.graph.edges[("xi1", "eta1")], {"weight": 0.3}
        )
        self.assertDictEqual(
            self.demo_params.graph.edges[("eta1", "y1")], {"weight": 1.1}
        )
        self.assertDictEqual(
            self.demo_params.graph.edges[("eta1", "y2")], {"weight": 1.2}
        )
        self.assertDictEqual(
            self.demo_params.graph.edges[("eta1", "y3")], {"weight": 1.3}
        )
        self.assertDictEqual(
            self.demo_params.graph.edges[("eta1", "y4")], {"weight": 1.4}
        )
        self.assertDictEqual(
            self.demo_params.graph.edges[("eta1", "eta2")], {"weight": 0.1}
        )
        self.assertDictEqual(
            self.demo_params.graph.edges[("xi1", "eta2")], {"weight": 0.2}
        )
        self.assertDictEqual(
            self.demo_params.graph.edges[("eta2", "y5")], {"weight": 0.7}
        )
        self.assertDictEqual(
            self.demo_params.graph.edges[("eta2", "y6")], {"weight": 0.8}
        )
        self.assertDictEqual(
            self.demo_params.graph.edges[("eta2", "y7")], {"weight": 0.9}
        )
        self.assertDictEqual(
            self.demo_params.graph.edges[("eta2", "y8")], {"weight": 1.0}
        )

        self.assertDictEqual(
            self.demo_params.err_graph.edges[("y1", "y5")], {"weight": 1.5}
        )
        self.assertDictEqual(
            self.demo_params.err_graph.edges[("y2", "y6")], {"weight": 1.6}
        )
        self.assertDictEqual(
            self.demo_params.err_graph.edges[("y2", "y4")], {"weight": 1.9}
        )
        self.assertDictEqual(
            self.demo_params.err_graph.edges[("y3", "y7")], {"weight": 1.7}
        )
        self.assertDictEqual(
            self.demo_params.err_graph.edges[("y4", "y8")], {"weight": 1.8}
        )
        self.assertDictEqual(
            self.demo_params.err_graph.edges[("y6", "y8")], {"weight": 2.0}
        )

        self.assertDictEqual(self.demo_params.err_graph.nodes["y1"], {"weight": 2.1})
        self.assertDictEqual(self.demo_params.err_graph.nodes["y2"], {"weight": 2.2})
        self.assertDictEqual(self.demo_params.err_graph.nodes["y3"], {"weight": 2.3})
        self.assertDictEqual(self.demo_params.err_graph.nodes["y4"], {"weight": 2.4})
        self.assertDictEqual(self.demo_params.err_graph.nodes["y5"], {"weight": 2.5})
        self.assertDictEqual(self.demo_params.err_graph.nodes["y6"], {"weight": 2.6})
        self.assertDictEqual(self.demo_params.err_graph.nodes["y7"], {"weight": 2.7})
        self.assertDictEqual(self.demo_params.err_graph.nodes["y8"], {"weight": 2.8})
        self.assertDictEqual(self.demo_params.err_graph.nodes["x1"], {"weight": 3.1})
        self.assertDictEqual(self.demo_params.err_graph.nodes["x2"], {"weight": 3.2})
        self.assertDictEqual(self.demo_params.err_graph.nodes["x3"], {"weight": 3.3})
        self.assertDictEqual(self.demo_params.err_graph.nodes["eta1"], {"weight": 2.9})
        self.assertDictEqual(self.demo_params.err_graph.nodes["eta2"], {"weight": 3.0})

    def test_get_full_graph_struct(self):
        full_struct = self.union._get_full_graph_struct()
        self.assertFalse(
            set(full_struct.nodes())
            - set(
                [
                    "yrsmill",
                    "unionsen",
                    "age",
                    "laboract",
                    "deferenc",
                    ".yrsmill",
                    ".unionsen",
                    ".age",
                    ".laboract",
                    ".deferenc",
                    "..ageyrsmill",
                    "..yrsmillage",
                ]
            )
        )
        self.assertFalse(
            set(full_struct.edges())
            - set(
                [
                    ("yrsmill", "unionsen"),
                    ("age", "laboract"),
                    ("age", "deferenc"),
                    ("deferenc", "laboract"),
                    ("deferenc", "unionsen"),
                    ("laboract", "unionsen"),
                    (".yrsmill", "yrsmill"),
                    (".unionsen", "unionsen"),
                    (".age", "age"),
                    (".laboract", "laboract"),
                    (".deferenc", "deferenc"),
                    ("..ageyrsmill", ".age"),
                    ("..ageyrsmill", ".yrsmill"),
                    ("..yrsmillage", ".age"),
                    ("..yrsmillage", ".yrsmill"),
                ]
            )
        )

    def test_active_trail_nodes(self):
        demo_nodes = ["x1", "x2", "x3", "y1", "y2", "y3", "y4", "y5", "y6", "y7", "y8"]
        for node in demo_nodes:
            self.assertSetEqual(
                self.demo.active_trail_nodes(node, struct="full")[node], set(demo_nodes)
            )

        union_nodes = self.union.graph.nodes()
        active_trails = self.union.active_trail_nodes(list(union_nodes), struct="full")
        for node in union_nodes:
            self.assertSetEqual(active_trails[node], set(union_nodes))

        self.assertSetEqual(
            self.union.active_trail_nodes(
                "age", observed=["laboract", "deferenc", "unionsen"]
            )["age"],
            {"age", "yrsmill"},
        )

    def test_get_scaling_indicators(self):
        demo_scaling_indicators = self.demo.get_scaling_indicators()
        self.assertTrue(demo_scaling_indicators["eta1"] in ["y1", "y2", "y3", "y4"])
        self.assertTrue(demo_scaling_indicators["eta2"] in ["y5", "y6", "y7", "y8"])
        self.assertTrue(demo_scaling_indicators["xi1"] in ["x1", "x2", "x3"])

        union_scaling_indicators = self.union.get_scaling_indicators()
        self.assertDictEqual(union_scaling_indicators, dict())

        custom_scaling_indicators = self.custom.get_scaling_indicators()
        self.assertTrue(custom_scaling_indicators["xi1"] in ["x1", "x2", "y1", "y4"])
        self.assertTrue(custom_scaling_indicators["eta1"] in ["y2", "y3"])
        self.assertTrue(custom_scaling_indicators["eta2"] in ["y5"])

    def test_to_lisrel(self):
        demo = SEMGraph(
            ebunch=[
                ("xi1", "x1", 1.000),
                ("xi1", "x2", 2.180),
                ("xi1", "x3", 1.819),
                ("xi1", "eta1", 1.483),
                ("eta1", "y1", 1.000),
                ("eta1", "y2", 1.257),
                ("eta1", "y3", 1.058),
                ("eta1", "y4", 1.265),
                ("eta1", "eta2", 0.837),
                ("xi1", "eta2", 0.572),
                ("eta2", "y5", 1.000),
                ("eta2", "y6", 1.186),
                ("eta2", "y7", 1.280),
                ("eta2", "y8", 1.266),
            ],
            latents=["xi1", "eta1", "eta2"],
            err_corr=[
                ("y1", "y5", 0.624),
                ("y2", "y6", 2.153),
                ("y2", "y4", 1.313),
                ("y3", "y7", 0.795),
                ("y4", "y8", 0.348),
                ("y6", "y8", 1.356),
            ],
            err_var={
                "x1": 0.082,
                "x2": 0.120,
                "x3": 0.467,
                "y1": 1.891,
                "y2": 7.373,
                "y3": 5.067,
                "y4": 3.148,
                "y5": 2.351,
                "y6": 4.954,
                "y7": 3.431,
                "y8": 3.254,
                "xi1": 0.448,
                "eta1": 3.956,
                "eta2": 0.172,
            },
        )
        demo_lisrel = demo.to_lisrel()

        indexing = []
        vars_ordered = [
            "y1",
            "y2",
            "y3",
            "y4",
            "y5",
            "y6",
            "y7",
            "y8",
            "x1",
            "x2",
            "x3",
            "xi1",
            "eta1",
            "eta2",
        ]
        for var in vars_ordered:
            indexing.append(demo_lisrel.eta.index(var))

        eta_reorder = [demo_lisrel.eta[i] for i in indexing]

        B_reorder = demo_lisrel.B[indexing, :][:, indexing]
        B_fixed_reorder = demo_lisrel.B_fixed_mask[indexing, :][:, indexing]

        zeta_reorder = demo_lisrel.zeta[indexing, :][:, indexing]
        zeta_fixed_reorder = demo_lisrel.zeta_fixed_mask[indexing, :][:, indexing]

        wedge_y_reorder = demo_lisrel.wedge_y[:, indexing]

        self.assertEqual(vars_ordered, eta_reorder)
        npt.assert_array_equal(
            B_reorder,
            np.array(
                [
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0],
                ]
            ),
        )

        npt.assert_array_equal(
            zeta_reorder,
            np.array(
                [
                    [1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 1, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0],
                    [0, 1, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
                    [1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 1, 0, 0, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0],
                    [0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 1, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
                ]
            ),
        )

        npt.assert_array_equal(
            B_fixed_reorder,
            np.array(
                [
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.000, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.257, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.058, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.265, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.000],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.186],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.280],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.266],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.000, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2.180, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.819, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1.483, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.572, 0.837, 0],
                ]
            ),
        )

        npt.assert_array_equal(
            zeta_fixed_reorder,
            np.array(
                [
                    [1.891, 0, 0, 0, 0.624, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 7.373, 0, 1.313, 0, 2.153, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 5.067, 0, 0, 0, 0.795, 0, 0, 0, 0, 0, 0, 0],
                    [0, 1.313, 0, 3.148, 0, 0, 0, 0.348, 0, 0, 0, 0, 0, 0],
                    [0.624, 0, 0, 0, 2.351, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 2.153, 0, 0, 0, 4.954, 0, 1.356, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0.795, 0, 0, 0, 3.431, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0.348, 0, 1.356, 0, 3.254, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0.082, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0.120, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.467, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.448, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 3.956, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0.172],
                ]
            ),
        )

        npt.assert_array_equal(
            demo_lisrel.wedge_y,
            np.array(
                [
                    [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0],
                ]
            ),
        )

    def test_to_from_lisrel(self):
        demo_lisrel = self.demo.to_lisrel()
        union_lisrel = self.union.to_lisrel()
        demo_params_lisrel = self.demo_params.to_lisrel()
        custom_lisrel = self.custom.to_lisrel()

        demo_graph = demo_lisrel.to_SEMGraph()
        union_graph = union_lisrel.to_SEMGraph()
        demo_params_graph = demo_params_lisrel.to_SEMGraph()
        custom_graph = custom_lisrel.to_SEMGraph()

        # Test demo
        self.assertSetEqual(set(self.demo.graph.nodes()), set(demo_graph.graph.nodes()))
        self.assertSetEqual(set(self.demo.graph.edges()), set(demo_graph.graph.edges()))

        self.assertSetEqual(
            set(self.demo.err_graph.nodes()), set(demo_graph.err_graph.nodes())
        )
        npt.assert_array_equal(
            nx.to_numpy_matrix(
                self.demo.err_graph, nodelist=sorted(self.demo.err_graph.nodes())
            ),
            nx.to_numpy_matrix(
                demo_graph.err_graph, nodelist=sorted(demo_graph.err_graph.nodes())
            ),
        )

        self.assertSetEqual(
            set(self.demo.full_graph_struct.nodes()),
            set(demo_graph.full_graph_struct.nodes()),
        )
        self.assertSetEqual(
            set(self.demo.full_graph_struct.edges()),
            set(demo_graph.full_graph_struct.edges()),
        )

        self.assertSetEqual(self.demo.latents, demo_graph.latents)
        self.assertSetEqual(self.demo.observed, demo_graph.observed)

        # Test union
        self.assertSetEqual(
            set(self.union.graph.nodes()), set(union_graph.graph.nodes())
        )
        self.assertSetEqual(
            set(self.union.graph.edges()), set(union_graph.graph.edges())
        )

        self.assertSetEqual(
            set(self.union.err_graph.nodes()), set(union_graph.err_graph.nodes())
        )
        npt.assert_array_equal(
            nx.to_numpy_matrix(
                self.union.err_graph, nodelist=sorted(self.union.err_graph.nodes())
            ),
            nx.to_numpy_matrix(
                union_graph.err_graph, nodelist=sorted(union_graph.err_graph.nodes())
            ),
        )

        self.assertSetEqual(
            set(self.union.full_graph_struct.nodes()),
            set(union_graph.full_graph_struct.nodes()),
        )
        self.assertSetEqual(
            set(self.union.full_graph_struct.edges()),
            set(union_graph.full_graph_struct.edges()),
        )

        self.assertSetEqual(self.union.latents, union_graph.latents)
        self.assertSetEqual(self.union.observed, union_graph.observed)

        # Test demo_params
        self.assertSetEqual(
            set(self.demo_params.graph.nodes()), set(demo_params_graph.graph.nodes())
        )
        self.assertSetEqual(
            set(self.demo_params.graph.edges()), set(demo_params_graph.graph.edges())
        )

        self.assertSetEqual(
            set(self.demo_params.err_graph.nodes()),
            set(demo_params_graph.err_graph.nodes()),
        )
        npt.assert_array_equal(
            nx.to_numpy_matrix(
                self.demo_params.err_graph,
                nodelist=sorted(self.demo_params.err_graph.nodes()),
                weight=None,
            ),
            nx.to_numpy_matrix(
                demo_graph.err_graph,
                nodelist=sorted(demo_params_graph.err_graph.nodes()),
                weight=None,
            ),
        )

        self.assertSetEqual(
            set(self.demo_params.full_graph_struct.nodes()),
            set(demo_params_graph.full_graph_struct.nodes()),
        )
        self.assertSetEqual(
            set(self.demo_params.full_graph_struct.edges()),
            set(demo_params_graph.full_graph_struct.edges()),
        )

        self.assertSetEqual(self.demo_params.latents, demo_params_graph.latents)
        self.assertSetEqual(self.demo_params.observed, demo_params_graph.observed)

        # Test demo
        self.assertSetEqual(
            set(self.custom.graph.nodes()), set(custom_graph.graph.nodes())
        )
        self.assertSetEqual(
            set(self.custom.graph.edges()), set(custom_graph.graph.edges())
        )

        self.assertSetEqual(
            set(self.custom.err_graph.nodes()), set(custom_graph.err_graph.nodes())
        )
        npt.assert_array_equal(
            nx.to_numpy_matrix(
                self.custom.err_graph, nodelist=sorted(self.custom.err_graph.nodes())
            ),
            nx.to_numpy_matrix(
                custom_graph.err_graph, nodelist=sorted(custom_graph.err_graph.nodes())
            ),
        )

        self.assertSetEqual(
            set(self.custom.full_graph_struct.nodes()),
            set(custom_graph.full_graph_struct.nodes()),
        )
        self.assertSetEqual(
            set(self.custom.full_graph_struct.edges()),
            set(custom_graph.full_graph_struct.edges()),
        )

        self.assertSetEqual(self.custom.latents, custom_graph.latents)
        self.assertSetEqual(self.custom.observed, custom_graph.observed)

    def test_iv_transformations_demo(self):
        scale = {"eta1": "y1", "eta2": "y5", "xi1": "x1"}

        self.assertRaises(ValueError, self.demo._iv_transformations, "x1", "y1", scale)

        for y in ["y2", "y3", "y4"]:
            full_graph, dependent_var = self.demo._iv_transformations(
                X="eta1", Y=y, scaling_indicators=scale
            )
            self.assertEqual(dependent_var, y)
            self.assertTrue((".y1", y) in full_graph.edges)
            self.assertFalse(("eta1", y) in full_graph.edges)

        for y in ["y6", "y7", "y8"]:
            full_graph, dependent_var = self.demo._iv_transformations(
                X="eta2", Y=y, scaling_indicators=scale
            )
            self.assertEqual(dependent_var, y)
            self.assertTrue((".y5", y) in full_graph.edges)
            self.assertFalse(("eta2", y) in full_graph.edges)

        full_graph, dependent_var = self.demo._iv_transformations(
            X="xi1", Y="eta1", scaling_indicators=scale
        )
        self.assertEqual(dependent_var, "y1")
        self.assertTrue((".eta1", "y1") in full_graph.edges())
        self.assertTrue((".x1", "y1") in full_graph.edges())
        self.assertFalse(("xi1", "eta1") in full_graph.edges())

        full_graph, dependent_var = self.demo._iv_transformations(
            X="xi1", Y="eta2", scaling_indicators=scale
        )
        self.assertEqual(dependent_var, "y5")
        self.assertTrue((".y1", "y5") in full_graph.edges())
        self.assertTrue((".eta2", "y5") in full_graph.edges())
        self.assertTrue((".x1", "y5") in full_graph.edges())
        self.assertFalse(("eta1", "eta2") in full_graph.edges())
        self.assertFalse(("xi1", "eta2") in full_graph.edges())

        full_graph, dependent_var = self.demo._iv_transformations(
            X="eta1", Y="eta2", scaling_indicators=scale
        )
        self.assertEqual(dependent_var, "y5")
        self.assertTrue((".y1", "y5") in full_graph.edges())
        self.assertTrue((".eta2", "y5") in full_graph.edges())
        self.assertTrue((".x1", "y5") in full_graph.edges())
        self.assertFalse(("eta1", "eta2") in full_graph.edges())
        self.assertFalse(("xi1", "eta2") in full_graph.edges())

    def test_iv_transformations_union(self):
        scale = {}
        for u, v in self.union.graph.edges():
            full_graph, dependent_var = self.union._iv_transformations(
                u, v, scaling_indicators=scale
            )
            self.assertFalse((u, v) in full_graph.edges())
            self.assertEqual(dependent_var, v)

    def test_get_ivs_demo(self):
        scale = {"eta1": "y1", "eta2": "y5", "xi1": "x1"}

        self.assertSetEqual(
            self.demo.get_ivs("eta1", "y2", scaling_indicators=scale),
            {"x1", "x2", "x3", "y3", "y7", "y8"},
        )
        self.assertSetEqual(
            self.demo.get_ivs("eta1", "y3", scaling_indicators=scale),
            {"x1", "x2", "x3", "y2", "y4", "y6", "y8"},
        )
        self.assertSetEqual(
            self.demo.get_ivs("eta1", "y4", scaling_indicators=scale),
            {"x1", "x2", "x3", "y3", "y6", "y7"},
        )

        self.assertSetEqual(
            self.demo.get_ivs("eta2", "y6", scaling_indicators=scale),
            {"x1", "x2", "x3", "y3", "y4", "y7"},
        )
        self.assertSetEqual(
            self.demo.get_ivs("eta2", "y7", scaling_indicators=scale),
            {"x1", "x2", "x3", "y2", "y4", "y6", "y8"},
        )
        self.assertSetEqual(
            self.demo.get_ivs("eta2", "y8", scaling_indicators=scale),
            {"x1", "x2", "x3", "y2", "y3", "y7"},
        )

        self.assertSetEqual(
            self.demo.get_ivs("xi1", "x2", scaling_indicators=scale),
            {"x3", "y1", "y2", "y3", "y4", "y5", "y6", "y7", "y8"},
        )
        self.assertSetEqual(
            self.demo.get_ivs("xi1", "x3", scaling_indicators=scale),
            {"x2", "y1", "y2", "y3", "y4", "y5", "y6", "y7", "y8"},
        )

        self.assertSetEqual(
            self.demo.get_ivs("xi1", "eta1", scaling_indicators=scale), {"x2", "x3"}
        )
        self.assertSetEqual(
            self.demo.get_ivs("xi1", "eta2", scaling_indicators=scale),
            {"x2", "x3", "y2", "y3", "y4"},
        )
        self.assertSetEqual(
            self.demo.get_ivs("eta1", "eta2", scaling_indicators=scale),
            {"x2", "x3", "y2", "y3", "y4"},
        )

    def test_get_conditional_ivs_demo(self):
        scale = {"eta1": "y1", "eta2": "y5", "xi1": "x1"}

        self.assertEqual(
            self.demo.get_conditional_ivs("eta1", "y2", scaling_indicators=scale), []
        )
        self.assertEqual(
            self.demo.get_conditional_ivs("eta1", "y3", scaling_indicators=scale), []
        )
        self.assertEqual(
            self.demo.get_conditional_ivs("eta1", "y4", scaling_indicators=scale), []
        )

        self.assertEqual(
            self.demo.get_conditional_ivs("eta2", "y6", scaling_indicators=scale), []
        )
        self.assertEqual(
            self.demo.get_conditional_ivs("eta2", "y7", scaling_indicators=scale), []
        )
        self.assertEqual(
            self.demo.get_conditional_ivs("eta2", "y8", scaling_indicators=scale), []
        )

        self.assertEqual(
            self.demo.get_conditional_ivs("xi1", "x2", scaling_indicators=scale), []
        )
        self.assertEqual(
            self.demo.get_conditional_ivs("xi1", "x3", scaling_indicators=scale), []
        )

        self.assertEqual(
            self.demo.get_conditional_ivs("xi1", "eta1", scaling_indicators=scale), []
        )
        self.assertEqual(
            self.demo.get_conditional_ivs("xi1", "eta2", scaling_indicators=scale), []
        )
        self.assertEqual(
            self.demo.get_conditional_ivs("eta1", "eta2", scaling_indicators=scale), []
        )

    def test_get_ivs_union(self):
        scale = {}
        self.assertSetEqual(
            self.union.get_ivs("yrsmill", "unionsen", scaling_indicators=scale), set()
        )
        self.assertSetEqual(
            self.union.get_ivs("deferenc", "unionsen", scaling_indicators=scale), set()
        )
        self.assertSetEqual(
            self.union.get_ivs("laboract", "unionsen", scaling_indicators=scale), set()
        )
        self.assertSetEqual(
            self.union.get_ivs("deferenc", "laboract", scaling_indicators=scale), set()
        )
        self.assertSetEqual(
            self.union.get_ivs("age", "laboract", scaling_indicators=scale), {"yrsmill"}
        )
        self.assertSetEqual(
            self.union.get_ivs("age", "deferenc", scaling_indicators=scale), {"yrsmill"}
        )

    def test_get_conditional_ivs_union(self):
        self.assertEqual(
            self.union.get_conditional_ivs("yrsmill", "unionsen"),
            [("age", {"laboract", "deferenc"})],
        )
        # This case wouldn't have conditonal IV if the Total effect between `deferenc` and
        # `unionsen` needs to be computed because one of the conditional variable lies on the
        # effect path.
        self.assertEqual(
            self.union.get_conditional_ivs("deferenc", "unionsen"),
            [("age", {"yrsmill", "laboract"})],
        )
        self.assertEqual(
            self.union.get_conditional_ivs("laboract", "unionsen"),
            [("age", {"yrsmill", "deferenc"})],
        )
        self.assertEqual(self.union.get_conditional_ivs("deferenc", "laboract"), [])

        self.assertEqual(
            self.union.get_conditional_ivs("age", "laboract"),
            [("yrsmill", {"deferenc"})],
        )
        self.assertEqual(self.union.get_conditional_ivs("age", "deferenc"), [])

    def test_iv_transformations_custom(self):
        scale_custom = {"eta1": "y2", "eta2": "y5", "xi1": "x1"}

        full_graph, var = self.custom._iv_transformations(
            "xi1", "x2", scaling_indicators=scale_custom
        )
        self.assertEqual(var, "x2")
        self.assertTrue((".x1", "x2") in full_graph.edges())
        self.assertFalse(("xi1", "x2") in full_graph.edges())

        full_graph, var = self.custom._iv_transformations(
            "xi1", "y4", scaling_indicators=scale_custom
        )
        self.assertEqual(var, "y4")
        self.assertTrue((".x1", "y4") in full_graph.edges())
        self.assertFalse(("xi1", "y4") in full_graph.edges())

        full_graph, var = self.custom._iv_transformations(
            "xi1", "y1", scaling_indicators=scale_custom
        )
        self.assertEqual(var, "y1")
        self.assertTrue((".x1", "y1") in full_graph.edges())
        self.assertFalse(("xi1", "y1") in full_graph.edges())
        self.assertFalse(("y4", "y1") in full_graph.edges())

        full_graph, var = self.custom._iv_transformations(
            "xi1", "eta1", scaling_indicators=scale_custom
        )
        self.assertEqual(var, "y2")
        self.assertTrue((".eta1", "y2") in full_graph.edges())
        self.assertTrue((".x1", "y2") in full_graph.edges())
        self.assertFalse(("y1", "eta1") in full_graph.edges())
        self.assertFalse(("xi1", "eta1") in full_graph.edges())

        full_graph, var = self.custom._iv_transformations(
            "y1", "eta1", scaling_indicators=scale_custom
        )
        self.assertEqual(var, "y2")
        self.assertTrue((".eta1", "y2") in full_graph.edges())
        self.assertTrue((".x1", "y2") in full_graph.edges())
        self.assertFalse(("y1", "eta1") in full_graph.edges())
        self.assertFalse(("xi1", "eta1") in full_graph.edges())

        full_graph, var = self.custom._iv_transformations(
            "y1", "eta2", scaling_indicators=scale_custom
        )
        self.assertEqual(var, "y5")
        self.assertTrue((".eta2", "y5") in full_graph.edges())
        self.assertFalse(("y1", "eta2") in full_graph.edges())

        full_graph, var = self.custom._iv_transformations(
            "y4", "y1", scaling_indicators=scale_custom
        )
        self.assertEqual(var, "y1")
        self.assertFalse(("y4", "y1") in full_graph.edges())

        full_graph, var = self.custom._iv_transformations(
            "eta1", "y3", scaling_indicators=scale_custom
        )
        self.assertEqual(var, "y3")
        self.assertTrue((".y2", "y3") in full_graph.edges())
        self.assertFalse(("eta1", "y3") in full_graph.edges())

    def test_get_ivs_custom(self):
        scale_custom = {"eta1": "y2", "eta2": "y5", "xi1": "x1"}

        self.assertSetEqual(
            self.custom.get_ivs("xi1", "x2", scaling_indicators=scale_custom),
            {"y1", "y2", "y3", "y4", "y5"},
        )
        self.assertSetEqual(
            self.custom.get_ivs("xi1", "y4", scaling_indicators=scale_custom), {"x2"}
        )
        self.assertSetEqual(
            self.custom.get_ivs("xi1", "y1", scaling_indicators=scale_custom),
            {"x2", "y4"},
        )
        self.assertSetEqual(
            self.custom.get_ivs("xi1", "eta1", scaling_indicators=scale_custom),
            {"x2", "y4"},
        )
        # TODO: Test this and fix.
        self.assertSetEqual(
            self.custom.get_ivs("y1", "eta1", scaling_indicators=scale_custom),
            {"x2", "y4", "y5"},
        )
        self.assertSetEqual(
            self.custom.get_ivs("y1", "eta2", scaling_indicators=scale_custom),
            {"x1", "x2", "y2", "y3", "y4"},
        )
        self.assertSetEqual(
            self.custom.get_ivs("y4", "y1", scaling_indicators=scale_custom), set()
        )
        self.assertSetEqual(
            self.custom.get_ivs("eta1", "y3", scaling_indicators=scale_custom),
            {"x1", "x2", "y4"},
        )

    def test_small_model_ivs(self):
        model1 = SEMGraph(
            ebunch=[("X", "Y"), ("I", "X"), ("W", "I")],
            latents=[],
            err_corr=[("W", "Y")],
            err_var={},
        )
        self.assertEqual(model1.get_conditional_ivs("X", "Y"), [("I", {"W"})])

        model2 = SEMGraph(
            ebunch=[
                ("x", "y"),
                ("z", "x"),
                ("w", "z"),
                ("w", "u"),
                ("u", "x"),
                ("u", "y"),
            ],
            latents=["u"],
        )
        self.assertEqual(model2.get_conditional_ivs("x", "y"), [("z", {"w"})])

        model3 = SEMGraph(
            ebunch=[("x", "y"), ("u", "x"), ("u", "y"), ("z", "x")], latents=["u"]
        )
        self.assertEqual(model3.get_ivs("x", "y"), {"z"})

        model4 = SEMGraph(ebunch=[("x", "y"), ("z", "x"), ("u", "x"), ("u", "y")])
        self.assertEqual(model4.get_conditional_ivs("x", "y"), [("z", {"u"})])


class TestSEMAlg(unittest.TestCase):
    def setUp(self):
        self.demo = SEMGraph(
            ebunch=[
                ("xi1", "x1", 1.000),
                ("xi1", "x2", 2.180),
                ("xi1", "x3", 1.819),
                ("xi1", "eta1", 1.483),
                ("eta1", "y1", 1.000),
                ("eta1", "y2", 1.257),
                ("eta1", "y3", 1.058),
                ("eta1", "y4", 1.265),
                ("eta1", "eta2", 0.837),
                ("xi1", "eta2", 0.572),
                ("eta2", "y5", 1.000),
                ("eta2", "y6", 1.186),
                ("eta2", "y7", 1.280),
                ("eta2", "y8", 1.266),
            ],
            latents=["xi1", "eta1", "eta2"],
            err_corr=[
                ("y1", "y5", 0.624),
                ("y2", "y6", 2.153),
                ("y2", "y4", 1.313),
                ("y3", "y7", 0.795),
                ("y4", "y8", 0.348),
                ("y6", "y8", 1.356),
            ],
            err_var={
                "x1": 0.082,
                "x2": 0.120,
                "x3": 0.467,
                "y1": 1.891,
                "y2": 7.373,
                "y3": 5.067,
                "y4": 3.148,
                "y5": 2.351,
                "y6": 4.954,
                "y7": 3.431,
                "y8": 3.254,
                "xi1": 0.448,
                "eta1": 3.956,
                "eta2": 0.172,
            },
        )
        self.demo_lisrel = self.demo.to_lisrel()

        self.small_model = SEM.from_graph(
            ebunch=[("X", "Y", 0.3)], latents=[], err_var={"X": 0.1, "Y": 0.1}
        )
        self.small_model_lisrel = self.small_model.to_lisrel()

    def test_generate_samples(self):
        samples = self.small_model_lisrel.generate_samples(n_samples=100)
        samples = self.demo_lisrel.generate_samples(n_samples=100)



import itertools
from collections import defaultdict, namedtuple

import networkx as nx
import numpy as np
import pandas as pd
from tqdm.auto import tqdm

from pgmpy.factors import factor_product
from pgmpy.global_vars import SHOW_PROGRESS
from pgmpy.models import BayesianNetwork
from pgmpy.models import DynamicBayesianNetwork as DBN
from pgmpy.models import MarkovChain, MarkovNetwork
from pgmpy.sampling import BayesianModelInference, _return_samples
from pgmpy.utils.mathext import sample_discrete, sample_discrete_maps

State = namedtuple("State", ["var", "state"])


class BayesianModelSampling(BayesianModelInference):
    """
    Class for sampling methods specific to Bayesian Models

    Parameters
    ----------
    model: instance of BayesianNetwork
        model on which inference queries will be computed
    """

    def __init__(self, model):
        super(BayesianModelSampling, self).__init__(model)

    def forward_sample(
        self,
        size=1,
        include_latents=False,
        seed=None,
        show_progress=True,
        partial_samples=None,
    ):
        """
        Generates sample(s) from joint distribution of the bayesian network.

        Parameters
        ----------
        size: int
            size of sample to be generated

        include_latents: boolean
            Whether to include the latent variable values in the generated samples.

        seed: int (default: None)
            If a value is provided, sets the seed for numpy.random.

        show_progress: boolean
            Whether to show a progress bar of samples getting generated.

        partial_samples: pandas.DataFrame
            A pandas dataframe specifying samples on some of the variables in the model. If
            specified, the sampling procedure uses these sample values, instead of generating them.

        Returns
        -------
        sampled: pandas.DataFrame
            The generated samples

        Examples
        --------
        >>> from pgmpy.models import BayesianNetwork
        >>> from pgmpy.factors.discrete import TabularCPD
        >>> from pgmpy.sampling import BayesianModelSampling
        >>> student = BayesianNetwork([('diff', 'grade'), ('intel', 'grade')])
        >>> cpd_d = TabularCPD('diff', 2, [[0.6], [0.4]])
        >>> cpd_i = TabularCPD('intel', 2, [[0.7], [0.3]])
        >>> cpd_g = TabularCPD('grade', 3, [[0.3, 0.05, 0.9, 0.5], [0.4, 0.25,
        ...                0.08, 0.3], [0.3, 0.7, 0.02, 0.2]],
        ...                ['intel', 'diff'], [2, 2])
        >>> student.add_cpds(cpd_d, cpd_i, cpd_g)
        >>> inference = BayesianModelSampling(student)
        >>> inference.forward_sample(size=2)
        rec.array([(0, 0, 1), (1, 0, 2)], dtype=
                  [('diff', '<i8'), ('intel', '<i8'), ('grade', '<i8')])
        """
        sampled = pd.DataFrame(columns=list(self.model.nodes()))

        if show_progress and SHOW_PROGRESS:
            pbar = tqdm(self.topological_order)
        else:
            pbar = self.topological_order

        if seed is not None:
            np.random.seed(seed)

        for node in pbar:
            if show_progress and SHOW_PROGRESS:
                pbar.set_description(f"Generating for node: {node}")

            # If values specified in partial_samples, use them. Else generate the values.
            if (partial_samples is not None) and (node in partial_samples.columns):
                sampled[node] = partial_samples.loc[:, node].values
            else:
                cpd = self.model.get_cpds(node)
                states = range(self.cardinality[node])
                evidence = cpd.variables[:0:-1]
                if evidence:
                    evidence_values = np.vstack([sampled[i] for i in evidence])

                    state_to_index, index_to_weight = self.pre_compute_reduce_maps(
                        variable=node
                    )
                    unique, inverse = np.unique(
                        evidence_values.T, axis=0, return_inverse=True
                    )
                    weight_index = np.array([state_to_index[tuple(u)] for u in unique])[
                        inverse
                    ]
                    sampled[node] = sample_discrete_maps(
                        states, weight_index, index_to_weight, size
                    )
                else:
                    weights = cpd.values
                    sampled[node] = sample_discrete(states, weights, size)

        samples_df = _return_samples(sampled, self.state_names_map)
        if not include_latents:
            samples_df.drop(self.model.latents, axis=1, inplace=True)
        return samples_df

    def rejection_sample(
        self,
        evidence=[],
        size=1,
        include_latents=False,
        seed=None,
        show_progress=True,
        partial_samples=None,
    ):
        """
        Generates sample(s) from joint distribution of the bayesian network,
        given the evidence.

        Parameters
        ----------
        evidence: list of `pgmpy.factor.State` namedtuples
            None if no evidence

        size: int
            size of sample to be generated

        include_latents: boolean
            Whether to include the latent variable values in the generated samples.

        seed: int (default: None)
            If a value is provided, sets the seed for numpy.random.

        show_progress: boolean
            Whether to show a progress bar of samples getting generated.

        partial_samples: pandas.DataFrame
            A pandas dataframe specifying samples on some of the variables in the model. If
            specified, the sampling procedure uses these sample values, instead of generating them.

        Returns
        -------
        sampled: pandas.DataFrame
            The generated samples

        Examples
        --------
        >>> from pgmpy.models import BayesianNetwork
        >>> from pgmpy.factors.discrete import TabularCPD
        >>> from pgmpy.factors.discrete import State
        >>> from pgmpy.sampling import BayesianModelSampling
        >>> student = BayesianNetwork([('diff', 'grade'), ('intel', 'grade')])
        >>> cpd_d = TabularCPD('diff', 2, [[0.6], [0.4]])
        >>> cpd_i = TabularCPD('intel', 2, [[0.7], [0.3]])
        >>> cpd_g = TabularCPD('grade', 3, [[0.3, 0.05, 0.9, 0.5], [0.4, 0.25,
        ...                0.08, 0.3], [0.3, 0.7, 0.02, 0.2]],
        ...                ['intel', 'diff'], [2, 2])
        >>> student.add_cpds(cpd_d, cpd_i, cpd_g)
        >>> inference = BayesianModelSampling(student)
        >>> evidence = [State(var='diff', state=0)]
        >>> inference.rejection_sample(evidence=evidence, size=2, return_type='dataframe')
                intel       diff       grade
        0         0          0          1
        1         0          0          1
        """

        if seed is not None:
            np.random.seed(seed)

        # If no evidence is given, it is equivalent to forward sampling.
        if len(evidence) == 0:
            return self.forward_sample(size=size, include_latents=include_latents)

        # Setup array to be returned
        sampled = pd.DataFrame()
        prob = 1
        i = 0

        # Do the sampling by generating samples from forward sampling and rejecting the
        # samples which do not match our evidence. Keep doing until we have enough
        # samples.
        if show_progress and SHOW_PROGRESS:
            pbar = tqdm(total=size)

        while i < size:
            _size = int(((size - i) / prob) * 1.5)

            # If partial_samples is specified, can only generate < partial_samples.shape[0] number of samples
            # at a time. For simplicity, just generate the same size as partial_samples.shape[0].
            if partial_samples is not None:
                _size = partial_samples.shape[0]

            _sampled = self.forward_sample(
                size=_size,
                include_latents=True,
                show_progress=False,
                partial_samples=partial_samples,
            )

            for var, state in evidence:
                _sampled = _sampled[_sampled[var] == state]

            prob = max(len(_sampled) / _size, 0.01)
            sampled = pd.concat([sampled, _sampled], axis=0, join="outer").iloc[
                :size, :
            ]
            i += _sampled.shape[0]

            if show_progress and SHOW_PROGRESS:
                # Update at maximum to `size`
                comp = _sampled.shape[0] if i < size else size - (i - _sampled.shape[0])
                pbar.update(comp)

        if show_progress and SHOW_PROGRESS:
            pbar.close()

        sampled = sampled.reset_index(drop=True)
        if not include_latents:
            sampled.drop(self.model.latents, axis=1, inplace=True)
        return sampled

    def likelihood_weighted_sample(
        self, evidence=[], size=1, include_latents=False, seed=None, show_progress=True
    ):
        """
        Generates weighted sample(s) from joint distribution of the bayesian
        network, that comply with the given evidence.
        'Probabilistic Graphical Model Principles and Techniques', Koller and
        Friedman, Algorithm 12.2 pp 493.

        Parameters
        ----------
        evidence: list of `pgmpy.factor.State` namedtuples
            None if no evidence

        size: int
            size of sample to be generated

        include_latents: boolean
            Whether to include the latent variable values in the generated samples.

        seed: int (default: None)
            If a value is provided, sets the seed for numpy.random.

        show_progress: boolean
            Whether to show a progress bar of samples getting generated.

        Returns
        -------
        sampled: A pandas.DataFrame
            The generated samples with corresponding weights

        Examples
        --------
        >>> from pgmpy.factors.discrete import State
        >>> from pgmpy.models import BayesianNetwork
        >>> from pgmpy.factors.discrete import TabularCPD
        >>> from pgmpy.sampling import BayesianModelSampling
        >>> student = BayesianNetwork([('diff', 'grade'), ('intel', 'grade')])
        >>> cpd_d = TabularCPD('diff', 2, [[0.6], [0.4]])
        >>> cpd_i = TabularCPD('intel', 2, [[0.7], [0.3]])
        >>> cpd_g = TabularCPD('grade', 3, [[0.3, 0.05, 0.9, 0.5], [0.4, 0.25,
        ...         0.08, 0.3], [0.3, 0.7, 0.02, 0.2]],
        ...         ['intel', 'diff'], [2, 2])
        >>> student.add_cpds(cpd_d, cpd_i, cpd_g)
        >>> inference = BayesianModelSampling(student)
        >>> evidence = [State('diff', 0)]
        >>> inference.likelihood_weighted_sample(evidence=evidence, size=2, return_type='recarray')
        rec.array([(0, 0, 1, 0.6), (0, 0, 2, 0.6)], dtype=
                  [('diff', '<i8'), ('intel', '<i8'), ('grade', '<i8'), ('_weight', '<f8')])
        """
        if seed is not None:
            np.random.seed(seed)

        # Convert evidence state names to number
        evidence = [
            (var, self.model.get_cpds(var).get_state_no(var, state))
            for var, state in evidence
        ]

        # Prepare the return dataframe
        sampled = pd.DataFrame(columns=list(self.model.nodes()))
        sampled["_weight"] = np.ones(size)
        evidence_dict = dict(evidence)

        if show_progress and SHOW_PROGRESS:
            pbar = tqdm(self.topological_order)
        else:
            pbar = self.topological_order

        # Do the sampling
        for node in pbar:
            if show_progress and SHOW_PROGRESS:
                pbar.set_description(f"Generating for node: {node}")

            cpd = self.model.get_cpds(node)
            states = range(self.cardinality[node])
            evidence = cpd.get_evidence()

            if evidence:
                evidence_values = np.vstack([sampled[i] for i in evidence])

                state_to_index, index_to_weight = self.pre_compute_reduce_maps(
                    variable=node
                )
                unique, inverse = np.unique(
                    evidence_values.T, axis=0, return_inverse=True
                )
                weight_index = np.array([state_to_index[tuple(u)] for u in unique])[
                    inverse
                ]

                if node in evidence_dict:
                    evidence_value = evidence_dict[node]
                    sampled[node] = evidence_value
                    sampled.loc[:, "_weight"] *= np.array(
                        list(
                            map(
                                lambda i: index_to_weight[weight_index[i]][
                                    evidence_value
                                ],
                                range(size),
                            )
                        )
                    )
                else:
                    sampled[node] = sample_discrete_maps(
                        states, weight_index, index_to_weight, size
                    )
            else:
                if node in evidence_dict:
                    sampled[node] = evidence_dict[node]
                    sampled.loc[:, "_weight"] *= np.array(
                        list(
                            map(lambda _: cpd.values[evidence_dict[node]], range(size))
                        )
                    )
                else:
                    sampled[node] = sample_discrete(states, cpd.values, size)

        # Postprocess the samples: Change state numbers to names, remove latents.
        samples_df = _return_samples(sampled, self.state_names_map)
        if not include_latents:
            samples_df.drop(self.model.latents, axis=1, inplace=True)
        return samples_df


class GibbsSampling(MarkovChain):
    """
    Class for performing Gibbs sampling.

    Parameters
    ----------
    model: BayesianNetwork or MarkovNetwork
        Model from which variables are inherited and transition probabilities computed.

    Examples
    --------
    Initialization from a BayesianNetwork object:

    >>> from pgmpy.factors.discrete import TabularCPD
    >>> from pgmpy.models import BayesianNetwork
    >>> intel_cpd = TabularCPD('intel', 2, [[0.7], [0.3]])
    >>> sat_cpd = TabularCPD('sat', 2, [[0.95, 0.2], [0.05, 0.8]], evidence=['intel'], evidence_card=[2])
    >>> student = BayesianNetwork()
    >>> student.add_nodes_from(['intel', 'sat'])
    >>> student.add_edge('intel', 'sat')
    >>> student.add_cpds(intel_cpd, sat_cpd)
    >>> from pgmpy.sampling import GibbsSampling
    >>> gibbs_chain = GibbsSampling(student)
    >>> gibbs_chain.sample(size=3)
       intel  sat
    0      0    0
    1      0    0
    2      1    1
    """

    def __init__(self, model=None):
        super(GibbsSampling, self).__init__()
        if isinstance(model, BayesianNetwork):
            self._get_kernel_from_bayesian_model(model)
        elif isinstance(model, MarkovNetwork):
            self._get_kernel_from_markov_model(model)

    def _get_kernel_from_bayesian_model(self, model):
        """
        Computes the Gibbs transition models from a Bayesian Network.
        'Probabilistic Graphical Model Principles and Techniques', Koller and
        Friedman, Section 12.3.3 pp 512-513.

        Parameters
        ----------
        model: BayesianNetwork
            The model from which probabilities will be computed.
        """
        self.variables = np.array(model.nodes())
        self.latents = model.latents
        self.cardinalities = {
            var: model.get_cpds(var).variable_card for var in self.variables
        }

        for var in self.variables:
            other_vars = [v for v in self.variables if var != v]
            other_cards = [self.cardinalities[v] for v in other_vars]
            cpds = [cpd for cpd in model.cpds if var in cpd.scope()]
            prod_cpd = factor_product(*cpds)
            kernel = {}
            scope = set(prod_cpd.scope())
            for tup in itertools.product(*[range(card) for card in other_cards]):
                states = [State(v, s) for v, s in zip(other_vars, tup) if v in scope]
                prod_cpd_reduced = prod_cpd.to_factor().reduce(states, inplace=False)
                kernel[tup] = prod_cpd_reduced.values / sum(prod_cpd_reduced.values)
            self.transition_models[var] = kernel

    def _get_kernel_from_markov_model(self, model):
        """
        Computes the Gibbs transition models from a Markov Network.
        'Probabilistic Graphical Model Principles and Techniques', Koller and
        Friedman, Section 12.3.3 pp 512-513.

        Parameters
        ----------
        model: MarkovNetwork
            The model from which probabilities will be computed.
        """
        self.variables = np.array(model.nodes())
        self.latents = model.latents
        factors_dict = {var: [] for var in self.variables}
        for factor in model.get_factors():
            for var in factor.scope():
                factors_dict[var].append(factor)

        # Take factor product
        factors_dict = {
            var: factor_product(*factors) if len(factors) > 1 else factors[0]
            for var, factors in factors_dict.items()
        }
        self.cardinalities = {
            var: factors_dict[var].get_cardinality([var])[var] for var in self.variables
        }

        for var in self.variables:
            other_vars = [v for v in self.variables if var != v]
            other_cards = [self.cardinalities[v] for v in other_vars]
            kernel = {}
            factor = factors_dict[var]
            scope = set(factor.scope())
            for tup in itertools.product(*[range(card) for card in other_cards]):
                states = [
                    State(first_var, s)
                    for first_var, s in zip(other_vars, tup)
                    if first_var in scope
                ]
                reduced_factor = factor.reduce(states, inplace=False)
                kernel[tup] = reduced_factor.values / sum(reduced_factor.values)
            self.transition_models[var] = kernel

    def sample(self, start_state=None, size=1, seed=None, include_latents=False):
        """
        Sample from the Markov Chain.

        Parameters
        ----------
        start_state: dict or array-like iterable
            Representing the starting states of the variables. If None is passed, a random start_state is chosen.

        size: int
            Number of samples to be generated.

        seed: int (default: None)
            If a value is provided, sets the seed for numpy.random.

        include_latents: boolean
            Whether to include the latent variable values in the generated samples.

        Returns
        -------
        sampled: pandas.DataFrame
            The generated samples

        Examples
        --------
        >>> from pgmpy.factors.discrete import DiscreteFactor
        >>> from pgmpy.sampling import GibbsSampling
        >>> from pgmpy.models import MarkovNetwork
        >>> model = MarkovNetwork([('A', 'B'), ('C', 'B')])
        >>> factor_ab = DiscreteFactor(['A', 'B'], [2, 2], [1, 2, 3, 4])
        >>> factor_cb = DiscreteFactor(['C', 'B'], [2, 2], [5, 6, 7, 8])
        >>> model.add_factors(factor_ab, factor_cb)
        >>> gibbs = GibbsSampling(model)
        >>> gibbs.sample(size=4, return_tupe='dataframe')
           A  B  C
        0  0  1  1
        1  1  0  0
        2  1  1  0
        3  1  1  1
        """
        if start_state is None and self.state is None:
            self.state = self.random_state()
        elif start_state is not None:
            self.set_start_state(start_state)

        if seed is not None:
            np.random.seed(seed)

        types = [(str(var_name), "int") for var_name in self.variables]
        sampled = np.zeros(size, dtype=types).view(np.recarray)
        sampled[0] = tuple(st for var, st in self.state)
        for i in tqdm(range(size - 1)):
            for j, (var, st) in enumerate(self.state):
                other_st = tuple(st for v, st in self.state if var != v)
                next_st = sample_discrete(
                    list(range(self.cardinalities[var])),
                    self.transition_models[var][other_st],
                )[0]
                self.state[j] = State(var, next_st)
            sampled[i + 1] = tuple(st for var, st in self.state)

        samples_df = _return_samples(sampled)
        if not include_latents:
            samples_df.drop(self.latents, axis=1, inplace=True)
        return samples_df

    def generate_sample(
        self, start_state=None, size=1, include_latents=False, seed=None
    ):
        """
        Generator version of self.sample

        Returns
        -------
        List of State namedtuples, representing the assignment to all variables of the model.

        Examples
        --------
        >>> from pgmpy.factors.discrete import DiscreteFactor
        >>> from pgmpy.sampling import GibbsSampling
        >>> from pgmpy.models import MarkovNetwork
        >>> model = MarkovNetwork([('A', 'B'), ('C', 'B')])
        >>> factor_ab = DiscreteFactor(['A', 'B'], [2, 2], [1, 2, 3, 4])
        >>> factor_cb = DiscreteFactor(['C', 'B'], [2, 2], [5, 6, 7, 8])
        >>> model.add_factors(factor_ab, factor_cb)
        >>> gibbs = GibbsSampling(model)
        >>> gen = gibbs.generate_sample(size=2)
        >>> [sample for sample in gen]
        [[State(var='C', state=1), State(var='B', state=1), State(var='A', state=0)],
         [State(var='C', state=0), State(var='B', state=1), State(var='A', state=1)]]
        """
        if seed is not None:
            np.random.seed(seed)

        if start_state is None and self.state is None:
            self.state = self.random_state()
        elif start_state is not None:
            self.set_start_state(start_state)

        for i in range(size):
            for j, (var, st) in enumerate(self.state):
                other_st = tuple(st for v, st in self.state if var != v)
                next_st = sample_discrete(
                    list(range(self.cardinalities[var])),
                    self.transition_models[var][other_st],
                )[0]
                self.state[j] = State(var, next_st)
            if include_latents:
                yield self.state[:]
            else:
                yield [s for s in self.state if i not in self.latents]



import numpy as np
import pandas as pd

from pgmpy.sampling import BayesianModelInference


class BayesianModelProbability(BayesianModelInference):
    """
    Class to calculate probability (pmf) values specific to Bayesian Models
    """

    def __init__(self, model):
        """
        Class to calculate probability (pmf) values specific to Bayesian Models

        Parameters
        ----------
        model: Bayesian Model
            model on which inference queries will be computed
        """
        super(BayesianModelProbability, self).__init__(model)

    def _log_probability_node(self, data, ordering, node):
        """
        Evaluate the log probability of each datapoint for a specific node.

        Internal function used by log_probability().

        Parameters
        ----------
        data: array_like, shape (n_samples, n_features)
            List of n_features-dimensional data points.  Each row
            corresponds to a single data point.

        ordering: list
            ordering of columns in data, used by the Bayesian model.
            default is topological ordering used by model.

        node: Bayesian Model Node
            node from the Bayesian network.

        Returns
        -------
        Log probability of node: np.array (n_samples,)
            The array of log(density) evaluations. These are normalized to be
            probability densities, so values will be low for high-dimensional
            data.
        """

        def vec_translate(a, my_dict):
            return np.vectorize(my_dict.__getitem__)(a)

        cpd = self.model.get_cpds(node)

        # variable to probe: data[n], where n is the node number
        current = cpd.variables[0]
        current_idx = ordering.index(current)
        current_val = data[:, current_idx]
        current_no = vec_translate(current_val, cpd.name_to_no[current])

        # conditional dependencies E of the probed variable
        evidence = cpd.variables[:0:-1]
        evidence_idx = [ordering.index(ev) for ev in evidence]
        evidence_val = data[:, evidence_idx]
        evidence_no = np.empty_like(evidence_val, dtype=int)
        for i, ev in enumerate(evidence):
            evidence_no[:, i] = vec_translate(evidence_val[:, i], cpd.name_to_no[ev])

        if evidence:
            # there are conditional dependencies E for data[n] for this node
            # Here we retrieve the array: p(x[n]|E). We do this for each x in data.
            # We pick the specific node value from the arrays below.

            state_to_index, index_to_weight = self.pre_compute_reduce_maps(
                variable=node
            )
            unique, inverse = np.unique(evidence_no, axis=0, return_inverse=True)
            weights = np.array(
                [index_to_weight[state_to_index[tuple(u)]] for u in unique]
            )[inverse]
        else:
            # there are NO conditional dependencies for this node
            # retrieve array: p(x[n]).  We do this for each x in data.
            # We pick the specific node value from the arrays below.
            weights = np.array([cpd.values] * len(data))

        # pick the specific node value x[n] from the array p(x[n]|E) or p(x[n])
        # We do this for each x in data.
        probability_node = np.array([weights[i][cn] for i, cn in enumerate(current_no)])

        return np.log(probability_node)

    def log_probability(self, data, ordering=None):
        """
        Evaluate the logarithmic probability of each point in a data set.

        Parameters
        ----------
        data: pandas dataframe OR array_like, shape (n_samples, n_features)
            List of n_features-dimensional data points.  Each row
            corresponds to a single data point.

        ordering: list
            ordering of columns in data, used by the Bayesian model.
            default is topological ordering used by model.

        Returns
        -------
        Log probability of each datapoint: np.array (n_samples,)
            The array of log(density) evaluations. These are normalized to be
            probability densities, so values will be low for high-dimensional
            data.
        """
        if isinstance(data, pd.DataFrame):
            # use numpy array from now on.
            ordering = data.columns.to_list()
            data = data.values
        if ordering is None:
            ordering = self.topological_order
            data = data.loc[:, ordering].values

        logp = np.array(
            [self._log_probability_node(data, ordering, node) for node in ordering]
        )
        return np.sum(logp, axis=0)

    def score(self, data, ordering=None):
        """
        Compute the total log probability density under the model.

        Parameters
        ----------
        data: pandas dataframe OR array_like, shape (n_samples, n_features)
            List of n_features-dimensional data points.  Each row
            corresponds to a single data point.

        ordering: list
            ordering of columns in data, used by the Bayesian model.
            default is topological ordering used by model.

        Returns
        -------
        Log-likelihood of data: float
            This is normalized to be a probability density, so the value
            will be low for high-dimensional data.
        """
        return np.sum(self.log_probability(data, ordering))



import warnings
from itertools import chain, product
from math import log

import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from tqdm.auto import tqdm

from pgmpy.estimators import MaximumLikelihoodEstimator, ParameterEstimator
from pgmpy.factors.discrete import TabularCPD
from pgmpy.global_vars import SHOW_PROGRESS
from pgmpy.models import BayesianNetwork


class ExpectationMaximization(ParameterEstimator):
    def __init__(self, model, data, **kwargs):
        """
        Class used to compute parameters for a model using Expectation
        Maximization (EM).

        EM is an iterative algorithm commonly used for
        estimation in the case when there are latent variables in the model.
        The algorithm iteratively improves the parameter estimates maximizing
        the likelihood of the given data.

        Parameters
        ----------
        model: A pgmpy.models.BayesianNetwork instance

        data: pandas DataFrame object
            DataFrame object with column names identical to the variable names
            of the network.  (If some values in the data are missing the data
            cells should be set to `numpy.NaN`.  Note that pandas converts each
            column containing `numpy.NaN`s to dtype `float`.)

        state_names: dict (optional)
            A dict indicating, for each variable, the discrete set of states
            that the variable can take. If unspecified, the observed values in
            the data set are taken to be the only possible states.

        complete_samples_only: bool (optional, default `True`)
            Specifies how to deal with missing data, if present. If set to
            `True` all rows that contain `np.NaN` somewhere are ignored. If
            `False` then, for each variable, every row where neither the
            variable nor its parents are `np.NaN` is used.

        Examples
        --------
        >>> import numpy as np
        >>> import pandas as pd
        >>> from pgmpy.models import BayesianNetwork
        >>> from pgmpy.estimators import ExpectationMaximization
        >>> data = pd.DataFrame(np.random.randint(low=0, high=2, size=(1000, 5)),
        ...                       columns=['A', 'B', 'C', 'D', 'E'])
        >>> model = BayesianNetwork([('A', 'B'), ('C', 'B'), ('C', 'D'), ('B', 'E')])
        >>> estimator = ExpectationMaximization(model, data)
        """
        if not isinstance(model, BayesianNetwork):
            raise NotImplementedError(
                "Expectation Maximization is only implemented for BayesianNetwork"
            )

        super(ExpectationMaximization, self).__init__(model, data, **kwargs)
        self.model_copy = self.model.copy()

    def _get_log_likelihood(self, datapoint):
        """
        Computes the likelihood of a given datapoint. Goes through each
        CPD matching the combination of states to get the value and multiplies
        them together.
        """
        likelihood = 0
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            for cpd in self.model_copy.cpds:
                scope = set(cpd.scope())
                likelihood += log(
                    cpd.get_value(
                        **{
                            key: value
                            for key, value in datapoint.items()
                            if key in scope
                        }
                    )
                )
        return likelihood

    def _parallel_compute_weights(
        self, data_unique, latent_card, n_counts, offset, batch_size
    ):
        cache = []

        for i in range(offset, min(offset + batch_size, data_unique.shape[0])):
            v = list(product(*[range(card) for card in latent_card.values()]))
            latent_combinations = np.array(v, dtype=int)
            df = data_unique.iloc[[i] * latent_combinations.shape[0]].reset_index(
                drop=True
            )
            for index, latent_var in enumerate(latent_card.keys()):
                df[latent_var] = latent_combinations[:, index]
            weights = df.apply(lambda t: self._get_log_likelihood(dict(t)), axis=1)
            df["_weight"] = (weights / weights.sum()) * n_counts[
                tuple(data_unique.iloc[i])
            ]
            cache.append(df)

        return pd.concat(cache, copy=False)

    def _compute_weights(self, n_jobs, latent_card, batch_size):
        """
        For each data point, creates extra data points for each possible combination
        of states of latent variables and assigns weights to each of them.
        """

        data_unique = self.data.drop_duplicates()
        n_counts = self.data.groupby(list(self.data.columns)).size().to_dict()

        cache = Parallel(n_jobs=n_jobs)(
            delayed(self._parallel_compute_weights)(
                data_unique, latent_card, n_counts, i, batch_size
            )
            for i in range(0, data_unique.shape[0], batch_size)
        )

        return pd.concat(cache, copy=False)

    def _is_converged(self, new_cpds, atol=1e-08):
        """
        Checks if the values of `new_cpds` is within tolerance limits of current
        model cpds.
        """
        for cpd in new_cpds:
            if not cpd.__eq__(self.model_copy.get_cpds(node=cpd.scope()[0]), atol=atol):
                return False
        return True

    def get_parameters(
        self,
        latent_card=None,
        max_iter=100,
        atol=1e-08,
        n_jobs=-1,
        batch_size=1000,
        seed=None,
        show_progress=True,
    ):
        """
        Method to estimate all model parameters (CPDs) using Expecation Maximization.

        Parameters
        ----------
        latent_card: dict (default: None)
            A dictionary of the form {latent_var: cardinality} specifying the
            cardinality (number of states) of each latent variable. If None,
            assumes `2` states for each latent variable.

        max_iter: int (default: 100)
            The maximum number of iterations the algorithm is allowed to run for.
            If max_iter is reached, return the last value of parameters.

        atol: int (default: 1e-08)
            The absolute accepted tolerance for checking convergence. If the parameters
            change is less than atol in an iteration, the algorithm will exit.

        n_jobs: int (default: -1)
            Number of jobs to run in parallel. Default: -1 uses all the processors.
            Suggest to use n_jobs=1 when dataset size is less than 1000.

        batch_size: int (default: 1000)
            Number of data used to compute weights in a batch.

        seed: int
            The random seed to use for generating the intial values.

        show_progress: boolean (default: True)
            Whether to show a progress bar for iterations.

        Returns
        -------
        Estimated paramters (CPDs): list
            A list of estimated CPDs for the model.

        Examples
        --------
        >>> import numpy as np
        >>> import pandas as pd
        >>> from pgmpy.models import BayesianNetwork
        >>> from pgmpy.estimators import ExpectationMaximization as EM
        >>> data = pd.DataFrame(np.random.randint(low=0, high=2, size=(1000, 3)),
        ...                       columns=['A', 'C', 'D'])
        >>> model = BayesianNetwork([('A', 'B'), ('C', 'B'), ('C', 'D')], latents={'B'})
        >>> estimator = EM(model, data)
        >>> estimator.get_parameters(latent_card={'B': 3})
        [<TabularCPD representing P(C:2) at 0x7f7b534251d0>,
        <TabularCPD representing P(B:3 | C:2, A:2) at 0x7f7b4dfd4da0>,
        <TabularCPD representing P(A:2) at 0x7f7b4dfd4fd0>,
        <TabularCPD representing P(D:2 | C:2) at 0x7f7b4df822b0>]
        """
        # Step 1: Parameter checks
        if latent_card is None:
            latent_card = {var: 2 for var in self.model_copy.latents}

        # Step 2: Create structures/variables to be used later.
        n_states_dict = {key: len(value) for key, value in self.state_names.items()}
        n_states_dict.update(latent_card)
        for var in self.model_copy.latents:
            self.state_names[var] = list(range(n_states_dict[var]))

        # Step 3: Initialize random CPDs if starting values aren't provided.
        if seed is not None:
            np.random.seed(seed)

        cpds = []
        for node in self.model_copy.nodes():
            parents = list(self.model_copy.predecessors(node))
            cpds.append(
                TabularCPD.get_random(
                    variable=node,
                    evidence=parents,
                    cardinality={
                        var: n_states_dict[var] for var in chain([node], parents)
                    },
                    state_names={
                        var: self.state_names[var] for var in chain([node], parents)
                    },
                )
            )

        self.model_copy.add_cpds(*cpds)

        if show_progress and SHOW_PROGRESS:
            pbar = tqdm(total=max_iter)

        # Step 4: Run the EM algorithm.
        for _ in range(max_iter):
            # Step 4.1: E-step: Expands the dataset and computes the likelihood of each
            #           possible state of latent variables.
            weighted_data = self._compute_weights(n_jobs, latent_card, batch_size)
            # Step 4.2: M-step: Uses the weights of the dataset to do a weighted MLE.
            new_cpds = MaximumLikelihoodEstimator(
                self.model_copy, weighted_data
            ).get_parameters(n_jobs=n_jobs, weighted=True)

            # Step 4.3: Check of convergence and max_iter
            if self._is_converged(new_cpds, atol=atol):
                if show_progress and SHOW_PROGRESS:
                    pbar.close()
                return new_cpds

            else:
                self.model_copy.cpds = new_cpds
                if show_progress and SHOW_PROGRESS:
                    pbar.update(1)

        return cpds



import unittest

from pgmpy.utils import get_example_model
from pgmpy.factors.discrete import TabularCPD
from pgmpy.inference import ApproxInference, VariableElimination


class TestApproxInference(unittest.TestCase):
    def setUp(self):
        self.alarm_model = get_example_model("alarm")
        self.infer_alarm = ApproxInference(self.alarm_model)
        self.alarm_ve = VariableElimination(self.alarm_model)

    def test_query_marg(self):
        query_results = self.infer_alarm.query(variables=["HISTORY"])
        ve_results = self.alarm_ve.query(variables=["HISTORY"])
        self.assertTrue(query_results.__eq__(ve_results, atol=0.01))

        query_results = self.infer_alarm.query(variables=["HISTORY", "CVP"], joint=True)
        ve_results = self.alarm_ve.query(variables=["HISTORY", "CVP"], joint=True)
        self.assertTrue(query_results.__eq__(ve_results, atol=0.01))

        query_results = self.infer_alarm.query(
            variables=["HISTORY", "CVP"], joint=False
        )
        ve_results = self.alarm_ve.query(variables=["HISTORY", "CVP"], joint=False)
        for var in ["HISTORY", "CVP"]:
            self.assertTrue(query_results[var].__eq__(ve_results[var], atol=0.01))

    def test_query_evidence(self):
        query_results = self.infer_alarm.query(
            variables=["HISTORY"], evidence={"PVSAT": "LOW"}, joint=True
        )
        ve_results = self.alarm_ve.query(
            variables=["HISTORY"], evidence={"PVSAT": "LOW"}, joint=True
        )
        self.assertTrue(query_results.__eq__(ve_results, atol=0.01))

        query_results = self.infer_alarm.query(
            variables=["HISTORY", "CVP"], evidence={"PVSAT": "LOW"}, joint=True
        )
        ve_results = self.alarm_ve.query(
            variables=["HISTORY", "CVP"], evidence={"PVSAT": "LOW"}, joint=True
        )
        self.assertTrue(query_results.__eq__(ve_results, atol=0.01))

        query_results = self.infer_alarm.query(
            variables=["HISTORY", "CVP"], evidence={"PVSAT": "LOW"}, joint=False
        )
        ve_results = self.alarm_ve.query(
            variables=["HISTORY", "CVP"], evidence={"PVSAT": "LOW"}, joint=False
        )
        for var in ["HISTORY", "CVP"]:
            self.assertTrue(query_results[var].__eq__(ve_results[var], atol=0.01))

    def test_virtual_evidence(self):
        virtual_evid = TabularCPD(
            "PAP",
            3,
            [[0.2], [0.3], [0.5]],
            state_names={"PAP": ["LOW", "NORMAL", "HIGH"]},
        )
        query_results = self.infer_alarm.query(
            variables=["HISTORY"], virtual_evidence=[virtual_evid]
        )
        ve_results = self.alarm_ve.query(
            variables=["HISTORY"], virtual_evidence=[virtual_evid]
        )
        self.assertTrue(query_results.__eq__(ve_results, atol=0.01))

        query_results = self.infer_alarm.query(
            variables=["HISTORY"],
            evidence={"PVSAT": "LOW"},
            virtual_evidence=[virtual_evid],
            joint=True,
        )
        ve_results = self.alarm_ve.query(
            variables=["HISTORY"],
            evidence={"PVSAT": "LOW"},
            virtual_evidence=[virtual_evid],
            joint=True,
        )
        self.assertTrue(query_results.__eq__(ve_results, atol=0.01))
